"""
===========================================
Poisson Gamma Dynamical Systems
Aaron Schein, Hanna Wallach and Mingyuan Zhou
Published in Neural Information Processing Systems 2016

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import os
import copy
import time
import numpy as np

from ._basic_model import Basic_Model
from .._sampler import Basic_Sampler
from .._utils import *

class PGDS(Basic_Model):
    def __init__(self, K: int, device='gpu'):
        """
        The basic model for PGDS
        Inputs:
            K      : [int] number of topics in PGDS;
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(PGDS, self).__init__()
        setattr(self, '_model_name', 'PGDS')

        self._model_setting.K = K
        self._model_setting.L = 1
        self._model_setting.device = device

        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)


    def initial(self, data: np.ndarray):
        '''
        Inintial the parameters of PGDS with the input documents
        Inputs:
            data : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary length of V

        Attributes:
            @public:
                global_params.Phi  : [np.ndarray] V*K matrix, K topics with a vocabulary length of V
                local_params.Theta : [np.ndarray] N*K matrix, the topic propotions of N documents

            @private:
                _model_setting.V         : [int] scalar, the length of the vocabulary
                _model_setting.Stationary: [int] scalar,
                _hyper_params.Supara     : [dict] scalar,
                _hyper_params.Para       : [dict] scalar,

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.V = data.shape[0]
        self._model_setting.Stationary = 1

        self.global_params.Phi = np.random.rand(self._model_setting.V, self._model_setting.K)
        self.global_params.Phi = self.global_params.Phi / np.sum(self.global_params.Phi, axis=0)
        self.global_params.Pi = np.eye(self._model_setting.K)
        self.global_params.Xi = 1
        self.global_params.V = np.ones((self._model_setting.K, 1))
        self.global_params.beta = 1
        self.global_params.h = np.zeros((self._model_setting.K, self._model_setting.K))
        self.global_params.n = np.zeros((self._model_setting.K, 1))
        self.global_params.rou = np.zeros((self._model_setting.K, 1))

        self._hyper_params.tao0 = 1
        self._hyper_params.gamma0 = 100
        self._hyper_params.eta0 = 0.1
        self._hyper_params.epilson0 = 0.1

    def train(self, iter_all: int, data: np.ndarray, is_train: bool = True):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            data       : [np.ndarray] V*N_train matrix, N_train bag-of-words vectors with a vocabulary length of V
            is_train   : [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:
                local_params.Theta : [np.ndarray] N_train*K matrix, the topic propotions of N_train documents

            @private:
                _model_setting.T         : [int] scalar, the number of the documents in the corpus
                _model_setting.Burnin    : [int] scalar, the burunin iterations of sampling
                _model_setting.Collection: [int] scalar, the Collection iterations of sampling
                _hyper_params.Para       : [dict] scalar,
        Outputs:
                local_params  : [Params] the local parameters of the probabilistic model

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'

        self._model_setting.T = data.shape[1]

        self.local_params.Theta = np.ones((self._model_setting.K, self._model_setting.T)) / self._model_setting.K
        self.local_params.Zeta = np.zeros((self._model_setting.T + 1, 1))
        self.local_params.delta = np.ones((self._model_setting.T, 1))


        A_VK = np.zeros((self._model_setting.V, self._model_setting.K))
        A_KT = np.zeros((self._model_setting.K, self._model_setting.T))
        L_dotkt = np.zeros((self._model_setting.K, self._model_setting.T + 1))
        L_kdott = np.zeros((self._model_setting.K, self._model_setting.T + 1))


        for iter in range(iter_all):

            start_time = time.time()

            A_KT, A_VK = self._sampler.multi_aug(data, self.global_params.Phi, self.local_params.Theta)

            L_KK = np.zeros((self._model_setting.K, self._model_setting.K))
            for t in range(self._model_setting.T - 1, 0, -1):  # T-1 : 1
                tmp1 = A_KT[:, t:t + 1] + L_dotkt[:, t + 1:t + 2]
                L_kdott[:, t:t + 1] = self._sampler.crt(tmp1, self._hyper_params.tao0 * np.dot(self.global_params.Pi, self.local_params.Theta[:, t - 1:t]))
                L_dotkt[:, t:t + 1], tmp = self._sampler.multi_aug(L_kdott[:, t:t + 1], self.global_params.Pi, self.local_params.Theta[:, t - 1:t])
                L_KK = L_KK + tmp

            if is_train:
                # Sample Phi
                self.global_params.Phi = self.Sample_Pi(A_VK, self._hyper_params.eta0)

                # Sample Pi
                Piprior = np.dot(self.global_params.V, np.transpose(self.global_params.V))
                Piprior[np.arange(Piprior.shape[0]), np.arange(Piprior.shape[1])] = 0
                Piprior = Piprior + np.diag(np.reshape(self.global_params.Xi * self.global_params.V, [self.global_params.V.shape[0]]))
                self.global_params.Pi = self.Sample_Pi(L_KK, Piprior)

            # Calculate Zeta   
            if (self._model_setting.Stationary == 1):
                for t in range(self._model_setting.T - 1, -1, -1):
                    self.local_params.Zeta[t] = np.log(1 + self.local_params.Zeta[t + 1] + self.local_params.delta[t])

            # Sample Theta  checked
            for t in range(self._model_setting.T):
                if t == 0:
                    shape = A_KT[:, t:t + 1] + L_dotkt[:, t + 1:t + 2] + self._hyper_params.tao0 * self.global_params.V
                else:
                    shape = A_KT[:, t:t + 1] + L_dotkt[:, t + 1:t + 2] + self._hyper_params.tao0 * np.dot(self.global_params.Pi, self.local_params.Theta[:, t - 1:t])

                scale = self.local_params.delta[t] + self._hyper_params.tao0 + self._hyper_params.tao0 * self.local_params.Zeta[t + 1];
                self.local_params.Theta[:, t:t + 1] = self._sampler.gamma(shape) / scale

            if is_train:
                # Sample Beta check
                shape = self._hyper_params.epilson0 + self._hyper_params.gamma0
                scale = self._hyper_params.epilson0 + np.sum(self.global_params.V)
                self.global_params.beta = self._sampler.gamma(shape) / scale

                # Sample q checked
                a = np.sum(L_dotkt, axis=1, keepdims=1)
                a[a == 0] = 1e-10
                b = self.global_params.V * (self.global_params.Xi + np.repeat(np.sum(self.global_params.V), self._model_setting.K, axis=0).reshape([self._model_setting.K, 1]) - self.global_params.V)
                b[b == 0] = 1e-10  #
                self.global_params.q = np.maximum(self._sampler.beta(b, a), 0.00001)

                # Sample h checked
                for k1 in range(self._model_setting.K):
                    for k2 in range(self._model_setting.K):
                        self.global_params.h[k1:k1 + 1, k2:k2 + 1] = self._sampler.crt(L_KK[k1:k1 + 1, k2:k2 + 1], Piprior[k1:k1 + 1, k2:k2 + 1])

                # Sample Xi checked
                shape = self._hyper_params.gamma0 / self._model_setting.K + np.trace(self.global_params.h)
                scale = self.global_params.beta - np.dot(np.transpose(self.global_params.V), log_max(self.global_params.q))
                self.global_params.Xi = self._sampler.gamma(shape) / scale

                # Sample V # check
                for k in range(self._model_setting.K):
                    L_kdott[k:k + 1, 0:1] = self._sampler.crt(A_KT[k:k + 1, 0:1] + L_dotkt[k:k + 1, 1:2],
                                                            np.reshape(self._hyper_params.tao0 * self.global_params.V[k], (1, 1)))
                    self.global_params.n[k] = np.sum(self.global_params.h[k, :] + np.transpose(self.global_params.h[:, k])) \
                                               - self.global_params.h[k, k] + L_kdott[k:k + 1, 0:1]

                    self.global_params.rou[k] = - log_max(self.global_params.q[k]) * (self.global_params.Xi + np.sum(self.global_params.V) - self.global_params.V[k]) \
                                                 - np.dot(np.transpose(log_max(self.global_params.q)), self.global_params.V) \
                                                 + log_max(self.global_params.q[k]) * self.global_params.V[k] + self.local_params.Zeta[0]

                shape_top = self._hyper_params.gamma0 / self._model_setting.K + self.global_params.n
                scale_top = self.global_params.beta + self.global_params.rou
                self.global_params.V = self._sampler.gamma(shape_top) / scale_top

            # Likelihood
            deltatmp = np.zeros([self._model_setting.K, self._model_setting.T])
            deltatmp[0:self._model_setting.K - 1, :] = np.transpose(self.local_params.delta)
            Theta_hat = deltatmp * self.local_params.Theta
            lambd = np.dot(self.global_params.Phi, self.local_params.Theta)
            # lambd = np.dot(self._hyper_params.Para['Phi'], self._hyper_params.Para['Theta'])

            like = np.sum(data * np.log(lambd) - lambd) / self._model_setting.V / self._model_setting.T

            end_time = time.time()
            stages = 'Training' if is_train else 'Testing'
            print(f'{stages} Stage: ',
                  f'epoch {iter:3d} takes {end_time - start_time:.2f} seconds. Likelihood:{like:8.3f}')

        return copy.deepcopy(self.local_params)


    def test(self, iter_all: int, data: np.ndarray):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        local_params = self.train(iter_all, data, is_train=False)

        return local_params


    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/PGDS.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')
        
        
    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model.

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])

    def Sample_Pi(self, WSZS, Eta):
        Phi = self._sampler.gamma(WSZS + Eta)
        tmp = np.sum(Phi, axis=0)
        temp_dex = np.where(tmp > 0)
        temp_dex_no = np.where(tmp <= 0)
        Phi[:, temp_dex] = Phi[:, temp_dex] / (tmp[temp_dex] + realmin)
        Phi[:, temp_dex_no] = 0
        return Phi




"""
===========================================
Model Sampler implemented on GPU
===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu
# License: BSD-3-Clause

import pycuda.curandom as curandom
import pycuda.driver as drv
import pycuda.tools
import pycuda.autoinit
from pycuda import gpuarray
from pycuda.compiler import SourceModule
from pycuda.curandom import XORWOWRandomNumberGenerator

import numpy as np
realmin = 2.2e-10

cuda_generator = XORWOWRandomNumberGenerator()

mod = SourceModule("""

#include <stdio.h>

__device__ int cudarand(long long seed)
{
    if (seed == 0)
    {
        seed = 1;
    }
    long long temp=(48271 * seed + 0) % 2147483647;
    return temp;
}

#include <stdio.h>

__global__ void _conv_multi_augmentation(float* randomseed, int* para, int* row_index, int* col_index, int* n_index, float* value_index, float* W1_nk1, float* D1_k1, float* W1_nk1_Aug, float* D1_k1_Aug)
{   
    int K0         = para[0];
    int K1         = para[1];
    int K1_S1      = para[2];
    int K1_S2      = para[3];
    int K1_S3      = para[4];
    int K1_S4      = para[5];
    int word_total = para[6];

    const int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if ((idx < word_total))
    {
        int v1 = row_index[idx];                 // row_index
        int v2 = col_index[idx];                 // col_index
        int n  = n_index[idx];                   // file_index
        float value = value_index[idx];          
        int seed = randomseed[idx] * 2147483647;

        int word_k1_min = 0;
        int word_k1_max = 0;
        int word_k2_min = 0;
        int word_k2_max = 0;

        // word_k1
        if ((v1 - K1_S3 + 1) > 0)
            word_k1_min = v1 - K1_S3 + 1;
        else
            word_k1_min = 0;

        if (v1 > K1_S1 -1)
            word_k1_max = K1_S1 -1;
        else
            word_k1_max = v1;

        int l_word_k1 = word_k1_max - word_k1_min + 1;

        // word_k2
        if ((v2 - K1_S4 + 1) > 0)
            word_k2_min = v2 - K1_S4 + 1;
        else
            word_k2_min = 0;

        if (v2 > K1_S2 -1)
            word_k2_max = K1_S2 -1;
        else
            word_k2_max = v2;

        int l_word_k2 = word_k2_max - word_k2_min + 1;

        // N*K0*K1_V1*K1_V2 => N*K1*K1_S1*K1_S2, K0*K1*K1_S3*K1_S4

        float MultRate_sum = 0;

        for (int i = 0; i < K1; i++)
        {
            for (int k = 0; k < (l_word_k1); k++)
            {
                for (int j = 0; j < (l_word_k2); j++)
                {
                    int word_k1 = word_k1_min + k;
                    int word_k2 = word_k2_min + j;
                    int temp_a = (n) * K1 * K1_S1 * K1_S2 + (i) * K1_S1 * K1_S2 + word_k1 * K1_S2 + (word_k2);
                    int temp_b = (i) * K1_S3 * K1_S4 + (v1 - word_k1) * K1_S4 + (v2 - word_k2);

                    MultRate_sum = MultRate_sum + W1_nk1[temp_a] * D1_k1[temp_b];
                }
            }
        }

        if (MultRate_sum == 0) 
        {
            return;
        }

        for (int token = 0; token<value; token++)
        {
            float cumsum=0.0;
            seed = cudarand(seed);
            float probrnd = ((float)(seed) / 2147483647.0) * MultRate_sum;
            int flag=0;

            for (int i = 0; i < K1; i++)
            {
                for (int k = 0; k < (l_word_k1); k++)
                {
                    for (int j = 0; j < (l_word_k2); j++)
                    {
                        int word_k1 = word_k1_min + k;
                        int word_k2 = word_k2_min + j;
                        int temp_a = (n) * K1 * K1_S1 * K1_S2 + (i) * K1_S1 * K1_S2 + word_k1 * K1_S2 + (word_k2);
                        int temp_b = (i) * K1_S3 * K1_S4 + (v1 - word_k1) * K1_S4 + (v2 - word_k2);

                        float prob = W1_nk1[temp_a] * D1_k1[temp_b];
                        cumsum += prob;
                        if (cumsum>=probrnd)
                        {
                             atomicAdd(&W1_nk1_Aug[temp_a], 1.0);
                             atomicAdd(&D1_k1_Aug[temp_b], 1.0);
                             flag = 1;
                             break;        
                        }
                    }

                    if (flag==1) break;
                }

                if (flag==1) break;
            }
        }

    }
}

__global__ void Multi_Sampler(float* randomseed, int* Para, int* X_value, int* X_rows, int* X_cols, float* Phi, float* Theta, float* XVK, float* XKJ)    //
{
    const int V = Para[0];
    const int K = Para[1];
    const int J = Para[2];
    const int N = Para[3];
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int seed = randomseed[idx]*2147483647.0;

    if (idx < N)
    {
        float cumsum = 0.0;
        float sum=0.0;
        int v = X_rows[idx];
        int j = X_cols[idx];

        for (int k = 0; k < K; k++)
        {
            sum += Phi[v*K + k] * Theta[k*J + j];
        }

        for (int token = 0; token<X_value[idx]; token++)
        {
            int Embedding_K=K-1;
            float sumprob=0.0;
            seed = cudarand(seed);
            float probrnd = ((float)(seed) / 2147483647.0) * sum;

            for (int k = 0; k < K; k++)//data
            {
                cumsum = Phi[v*K + k] * Theta[k*J + j];
                if (sumprob+cumsum>=probrnd)
                {
                    Embedding_K=k;
                    break;
                }
                sumprob+=cumsum;
            }

            atomicAdd(&XVK[v*K + Embedding_K], 1);
            atomicAdd(&XKJ[Embedding_K*J + j], 1);
        }
    }
}

__global__ void Crt_Conv_Multi_Sampler(float* randomseed, int* para, int* n_index, int* k1_index, int* row_index, int* col_index, float* value_index, float* D2_k2, float* W2_nk2, float* Phi_Theta, float* D2_k2_Aug, float* W2_nk2_Aug)
{
    int N = para[0];
    int K1 = para[1];
    int K1_S1 = para[2];
    int K1_S2 = para[3];
    int K2 = para[4];
    int K2_S1 = para[5];
    int K2_S2 = para[6];
    int K2_S3 = para[7];
    int K2_S4 = para[8];
    int word_total = para[9];

    const int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if ((idx < word_total))
    {
        int n = n_index[idx];
        int k1 = k1_index[idx];
        int row = row_index[idx];
        int col = col_index[idx];
        float value = value_index[idx];

        // Crt: N*K1*K1_S1*K1_S2

        int seed = randomseed[idx] * 2147483647;
        int temp_a = n*K1*K1_S1*K1_S2 + k1*K1_S1*K1_S2 + row*K1_S1 + col;
        float sum = Phi_Theta[temp_a]; // Phi_Theta: N*K1*K1_S1*K1_S2
        int table = 0;
        int token = 0;

        if (value<0.5)
        {
            table = 0;
            return;
        }
        else
        {
            for ( token = 1, table = 1; token<value; token++)
            {
                seed = cudarand(seed);
                float probrnd = ((float)(seed) / 2147483647.0);
                if (probrnd <= sum / (sum + token))
                    table++;
            }
        }

        //W1_nk1: N*K1*K1_S1*K1_S2 => W2_nk2: N*K2*K2_S1*K2_S2  D2_k2: K1*K2*K2_S3*K2_S4

        int word_k1_min = 0;
        int word_k1_max = 0;
        int word_k2_min = 0;
        int word_k2_max = 0;

        // word_k1
        if ((row - K2_S3 + 1) > 0)
            word_k1_min = row - K2_S3 + 1;
        else
            word_k1_min = 0;

        if (row > K2_S1 -1)
            word_k1_max = K2_S1 -1;
        else
            word_k1_max = row;

        int l_word_k1 = word_k1_max - word_k1_min + 1;

        // word_k2
        if ((col - K2_S4 + 1) > 0)
            word_k2_min = col - K2_S4 + 1;
        else
            word_k2_min = 0;

        if (col > K2_S2 -1)
            word_k2_max = K2_S2 -1;
        else
            word_k2_max = col;

        int l_word_k2 = word_k2_max - word_k2_min + 1;

        float MultRate_sum = 0;

        for (int i = 0; i < K2; i++)
        {
            for (int k = 0; k < (l_word_k1); k++)
            {
                for (int j = 0; j < (l_word_k2); j++)
                {
                    int word_k1 = word_k1_min + k;
                    int word_k2 = word_k2_min + j;
                    int temp_a = (n) * K2 * K2_S1 * K2_S2 + (i) * K2_S1 * K2_S2 + word_k1 * K2_S2 + (word_k2);
                    int temp_b = k1 * K2 * K2_S3 * K2_S4 +  (i) * K2_S3 * K2_S4 + (row - word_k1) * K2_S4 + (col - word_k2);
                    MultRate_sum = MultRate_sum + W2_nk2[temp_a] * D2_k2[temp_b];
                }
            }
        }

        if (MultRate_sum == 0) 
        {
            return;
        }


        for (int token = 0; token<table; token++)
        {
            float cumsum=0.0;
            seed = cudarand(seed);
            float probrnd = ((float)(seed) / 2147483647.0) * MultRate_sum;
            int flag=0;

            for (int i = 0; i < K2; i++)
            {
                for (int k = 0; k < (l_word_k1); k++)
                {
                    for (int j = 0; j < (l_word_k2); j++)
                    {
                        int word_k1 = word_k1_min + k;
                        int word_k2 = word_k2_min + j;
                        int temp_a = (n) * K2 * K2_S1 * K2_S2 + (i) * K2_S1 * K2_S2 + word_k1 * K2_S2 + (word_k2);
                        int temp_b = k1 * K2 * K2_S3 * K2_S4 +  (i) * K2_S3 * K2_S4 + (row - word_k1) * K2_S4 + (col - word_k2);

                        float prob = W2_nk2[temp_a] * D2_k2[temp_b];
                        cumsum += prob;
                        if (cumsum>=probrnd)
                        {
                             atomicAdd(&W2_nk2_Aug[temp_a], 1.0);
                             atomicAdd(&D2_k2_Aug[temp_b], 1.0);
                             flag = 1;
                             break;        
                        }
                    }

                    if (flag==1) break;
                }

                if (flag==1) break;
            }
        }

    }
}

__global__ void Crt_Multi_Sampler(float *randomseed, int* Para, int* Xt_to_t1, float* Phi_t1, float* Theta_t1, float* Xt1_VK, float* Xt1_KJ)
{

    const int V = Para[0];
    const int K = Para[1];
    const int J = Para[2];
    const int N = Para[3];
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int seed = randomseed[idx] * 2147483647;

    if (idx < N && Xt_to_t1[idx] >= 0.5 )
    {
        float sum = 0.0;
        float cumsum = 0.0;
        int token, table;
        int v = idx / J;   // row first
        int j = idx % J;   // row first

        for (int k = 0; k<K; k++)
        {
            sum += Phi_t1[v*K + k] * Theta_t1[k*J + j]; // C index is different of Matlab
        }

        for (token = 1, table = 1; token<Xt_to_t1[idx]; token++)
        {
            seed = cudarand(seed);
            float probrnd = ((float)(seed) / 2147483647.0);
            if (probrnd <= sum / (sum + token))
                table++;
        }

        for (token = 0; token<table; token++)
        {
            int Embedding_K = K - 1;
            float sumprob = 0.0;
            seed = cudarand(seed);
            float probrnd = ((float)(seed) / 2147483647.0) * sum;

            for (int k = 0; k < K; k++)
            {
                cumsum = Phi_t1[v*K + k] * Theta_t1[k*J + j];
                if (sumprob + cumsum >= probrnd)
                {
                    Embedding_K = k;
                    break;
                }
                sumprob += cumsum;
            }

            atomicAdd(&Xt1_VK[v*K + Embedding_K], 1);
            atomicAdd(&Xt1_KJ[Embedding_K*J + j], 1);
        }
    }
}

__global__ void Sum_Pooling(int* para, float* M_nk, float* M_nk_pool, float* M_nk_w)
{
    int N = para[0];
    int K1 = para[1];
    int K1_S1 = para[2];
    int K1_S2 = para[3];
    int word_total = para[4];
    int stride = para[5];
    int K1_S2_pool = para[6];

    const int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if ((idx < word_total))
    {   
        int row_index = idx / K1_S2_pool;
        int remain = idx - row_index*K1_S2_pool;

        int idx_stride = stride;
        if (remain == K1_S2_pool-1)
        {
            idx_stride = K1_S2 - (K1_S2_pool-1)*stride;
        }

        //printf("%d, %d, %d", row_index, remain, idx_stride);

        for (int i=0; i<idx_stride; i++)
        {
            int temp_a = row_index*K1_S2 + remain*stride + i;
            atomicAdd(&M_nk_pool[idx], M_nk[temp_a] + 0.0000001); 
        }

        for (int i=0; i<idx_stride; i++)
        {            
            int temp_a = row_index*K1_S2 + remain*stride + i;
            float rate = (M_nk[temp_a] + 0.0000001) / (M_nk_pool[idx]);
            atomicAdd(&M_nk_w[temp_a], rate); 
        }
    }
}

__global__ void Up_Pooling(int* para, float* M_nk, float* M_nk_pool, float* M_nk_w)
{
    int N = para[0];
    int K1 = para[1];
    int K1_S1 = para[2];
    int K1_S2 = para[3];
    int word_total = para[4];
    int stride = para[5];
    int K1_S2_pool = para[6];

    const int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if ((idx < word_total))
    {   
        int row_index = idx / K1_S2_pool;
        int remain = idx - row_index*K1_S2_pool;

        int idx_stride = stride;
        if (remain == K1_S2_pool-1)
        {
            idx_stride = K1_S2 - (K1_S2_pool-1)*stride;
        }

        for (int i=0; i<idx_stride; i++)
        {
            int temp_a = row_index*K1_S2 + remain*stride + i;
            float rate = M_nk_pool[idx] * M_nk_w[temp_a];
            if (rate <= 0.0000001)
            {
                rate = 0.0000001;
            }

/*            if (M_nk_pool[idx] == 0)
            {
                printf("value error");
            }*/
            atomicAdd(&M_nk[temp_a], rate); 
        }
    }
}
 """)

class model_sampler_gpu(object):

    def __init__(self, system_type='Windows', seed=0):
        """
        The basic class for sampling distribution on cpu
        """
        super(model_sampler_gpu, self).__init__()

        self.system_type = system_type
        self.seed = seed

    def multi_aug(self, X_t, Phi_t, Theta_t, dtype='dense'):

        if dtype=='dense':

            [V, J] = X_t.shape  # 词表，文档，主题
            K = Theta_t.shape[0]
            [X_t_rows, X_t_cols] = np.where(X_t > 0.5)
            X_t_values = X_t[(X_t_rows, X_t_cols)]
            N = len(X_t_values)  # number of sample point
            Para = np.array([V, K, J, N], dtype=np.int32)  #

        X_t_values = np.array(X_t_values, dtype=np.int32)
        X_t_rows = np.array(X_t_rows, dtype=np.int32)
        X_t_cols = np.array(X_t_cols, dtype=np.int32)

        Xt_to_t1_t = np.zeros([K, J], dtype=np.float32, order='C')
        WSZS_t = np.zeros([V, K], dtype=np.float32, order='C')
        Phi_t = np.array(Phi_t, dtype=np.float32, order='C')
        Theta_t = np.array(Theta_t, dtype=np.float32, order='C')

        if N != 0:

            block_x = int(400)
            grid_x = int(np.floor(N / block_x) + 1)
            # print("block: {:<8.4f}, grid:{:<8.4f}".format(block_x, grid_x))

            randomseed = np.random.rand(N)
            randomseed = np.array(randomseed, dtype=np.float32, order='C')
            func = mod.get_function('Multi_Sampler')
            func(drv.In(randomseed), drv.In(Para), drv.In(X_t_values), drv.In(X_t_rows), drv.In(X_t_cols), drv.In(Phi_t),
                 drv.In(Theta_t), drv.InOut(WSZS_t), drv.InOut(Xt_to_t1_t),
                 grid=(grid_x, 1, 1), block=(block_x, 1, 1))

        return Xt_to_t1_t, WSZS_t

    def crt_multi_aug(self, Xt_to_t1_t, Phi_t1, Theta_t1, dtype='dense'):

        if dtype=='dense':

            [K_t, J] = Xt_to_t1_t.shape
            K_t1 = Theta_t1.shape[0]
            N = K_t*J
            Para = np.array([K_t, K_t1, J, N], dtype=np.int32)


        Xt_to_t1_t = np.array(Xt_to_t1_t, dtype=np.int32, order='C')
        Xt_to_t1_t1 = np.zeros([K_t1, J], dtype=np.float32, order='C')
        WSZS_t1 = np.zeros([K_t, K_t1], dtype=np.float32, order='C')
        Phi_t1 = np.array(Phi_t1, dtype=np.float32, order='C')
        Theta_t1 = np.array(Theta_t1, dtype=np.float32, order='C')

        if N!=0:

            block_x = int(400)
            grid_x = int(np.floor(N / block_x) + 1)

            randomseed = np.random.rand(N)
            randomseed = np.array(randomseed, dtype=np.float32, order='C')

            func = mod.get_function('Crt_Multi_Sampler')
            func(drv.In(randomseed), drv.In(Para), drv.In(Xt_to_t1_t), drv.In(Phi_t1),
                 drv.In(Theta_t1), drv.InOut(WSZS_t1), drv.InOut(Xt_to_t1_t1),
                 grid=(grid_x, 1), block=(block_x, 1, 1))

        return Xt_to_t1_t1, WSZS_t1

    def conv_multi_aug(self, row_index, col_index, n_index, value_index, D1_k1, W1_nk1):

        # check by chaojies
        [K1, K0, K1_S3, K1_S4] = D1_k1.shape  # K1*K0*K1_S3*K1_S4
        [N, K1, K1_S1, K1_S2] = W1_nk1.shape  # N*K1*K1_S1*K1_S2

        # 增广矩阵用于更新s12维度上增广
        X_rows = gpuarray.to_gpu(np.array(row_index, dtype=np.int32, order='C'))
        X_cols = gpuarray.to_gpu(np.array(col_index + 1, dtype=np.int32, order='C'))  # padding!!!
        X_file_index = gpuarray.to_gpu(np.array(n_index, dtype=np.int32, order='C'))
        X_value = gpuarray.to_gpu(np.array(value_index, dtype=np.float32, order='C'))
        word_total = int(X_rows.size)

        if word_total == 0:
            return np.zeros([N, K1, K1_S1, K1_S2]), np.zeros([K1, K0, K1_S3, K1_S4])
        else:
            W1_nk1 = gpuarray.to_gpu(np.array(W1_nk1, dtype=np.float32, order='C'))
            D1_k1 = gpuarray.to_gpu(np.array(np.swapaxes(D1_k1, 0, 1), dtype=np.float32, order='C'))  # K1*K0*K1_S3*K1_S4

            W1_nk1_Aug = gpuarray.zeros(W1_nk1.shape, dtype=np.float32, order='C')
            D1_k1_Aug = gpuarray.zeros(D1_k1.shape, dtype=np.float32, order='C')

            randomseed = cuda_generator.gen_uniform([word_total], dtype=np.float32)

            # 转化为GPU的输入形式
            fuc = mod.get_function("_conv_multi_augmentation")
            Batch_Para = gpuarray.to_gpu(np.array([K0, K1, K1_S1, K1_S2, K1_S3, K1_S4, word_total], dtype=np.int32, order='C'))

            block_x = int(500)
            grid_x = int(np.floor(word_total / block_x) + 1)

            fuc(randomseed, Batch_Para, X_rows, X_cols, X_file_index, X_value, W1_nk1, D1_k1,
                W1_nk1_Aug, D1_k1_Aug, grid=(grid_x, 1, 1), block=(block_x, 1, 1))  # 一般最多512个并行线程

            return W1_nk1_Aug.get(), np.swapaxes(D1_k1_Aug.get(), 0, 1)




"""
===========================================
WEDTM Demo
Inter and Intra Topic Structure Learning with Word Embeddings
He Zhao, Lan Du, Wray Buntine, Mingyaun Zhou
Published in International Council for Machinery Lubrication 2018

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause


import os
import copy
import time
import math
import numpy as np

from ._basic_model import Basic_Model
from .._sampler import Basic_Sampler
from .._utils import *

from scipy import sparse

class WEDTM(Basic_Model):
    def __init__(self, K: [list], device='gpu'):
        """
        The basic model for WEDTM
        Inputs:
            K      : [list] number of topics of each layer;
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(WEDTM, self).__init__()
        setattr(self, '_model_name', 'WEDTM')

        self._model_setting.K = K
        self._model_setting.T = len(K)
        self._model_setting.device = device
        
        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)


    def initial(self, data):
        '''
        Inintial the parameters of WEDTM with the input documents
        Inputs:
            data : [np.ndarray] or [scipy.sparse.csc.csc_matrix] V*N matrix, N bag-of-words vectors with a vocabulary length of V

        Attributes:
            @public:
                global_params.Phi  : [np.ndarray] V*K matrix, K topics with a vocabulary length of V
                local_params.Theta : [np.ndarray] N*K matrix, the topic propotions of N documents

            @private:
                _model_setting.V        : [int] scalar, the length of the vocabulary

        '''
        self._model_setting.V = data.shape[0]
        self.global_params.Phi = np.zeros((self._model_setting.K[0], self._model_setting.V)).astype(int)


    def train(self, embeddings: np.ndarray, S: int, iter_all: int, data: np.ndarray, is_train: bool = True):
        '''
        Inputs:
            embeddings     : [np.ndarray] V*D, word embedding of training words
            S              : [int] sub topics
            iter_all       : [np.ndarray] scalar, the iterations of gibbs sampling
            data           : [np.ndarray] V*N_train matrix, N_train bag-of-words vectors with a vocabulary length of V
            is_train       : [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:

                local_params.Theta : [np.ndarray] N_train*K matrix, the topic propotions of N_train documents

            @private:
                _model_setting.N         : [int] scalar, the number of the documents in the corpus
                _model_setting.Iteration : [int] scalar, the iterations of sampling

        Outputs:
                local_params  : [Params] the local parameters of the probabilistic model

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'

        self._model_setting.Iteration = [iter_all] * self._model_setting.T
        self._model_setting.N = data.shape[1]

        # initial local paramters
        self.local_params.Theta = np.zeros((self._model_setting.K[0], self._model_setting.N)).astype(int)

        # WS the trained words' word index
        # DS the trained words' doc index
        # ZS the trained words' random theme
        words_num = np.sum(data)
        WS = np.zeros(words_num).astype(int)
        DS = np.zeros(words_num).astype(int)
        wi, di = np.where(data)
        cc = data[wi, di]
        pos = 0
        for i in range(len(cc)):
            WS[pos:pos+cc[i]] = wi[i]
            DS[pos:pos+cc[i]] = di[i]
            pos = pos+cc[i]

        a0 = 0.01
        b0 = 0.01
        e0 = 1
        f0 = 1
        beta0 = 0.05

        # Add the default word embedding
        embeddings = np.insert(embeddings, embeddings.shape[1], values=np.ones(self._model_setting.V), axis=1)
        self.Theta = [[]] * self._model_setting.T
        c_j = [[]] * (self._model_setting.T + 1)
        for t in range(self._model_setting.T + 1):
            c_j[t] = np.ones((1, self._model_setting.N))

        self.Phi = [{}] * self._model_setting.T
        Xt_to_t1 = [[]] * self._model_setting.T
        WSZS = [[]] * self._model_setting.T
        paraGlobal = [{}] * self._model_setting.T
        # Initialise beta for t = 1
        beta1, self.beta_para = self._init_beta(self._model_setting.K[0], self._model_setting.V, S, embeddings, beta0)

        for Tcurrent in range(self._model_setting.T):
            if Tcurrent == 0:  # layer 1, initial params.
                ZS = np.random.randint(self._model_setting.K[Tcurrent], size=(len(DS)))  # theme of each words
                self.local_params.Theta = np.zeros((self._model_setting.K[Tcurrent], self._model_setting.N)).astype(int)  # Theta (K,N) distribution of theme
                for i in range(len(ZS)):
                    self.local_params.Theta[ZS[i], DS[i]] += 1
                if is_train:
                    self.global_params.Phi = np.zeros((self._model_setting.K[Tcurrent], self._model_setting.V)).astype(int)  # ZSWS Phi (K,V) distribution of words
                    for i in range(len(ZS)):
                        self.global_params.Phi[ZS[i], WS[i]] += 1
                WSZS[Tcurrent] = self.global_params.Phi.T
                Xt_to_t1[Tcurrent] = self.local_params.Theta
                n_dot_k = np.sum(self.local_params.Theta, 1)  # count number of each theme in doc
                p_j = self._calculate_pj(c_j, Tcurrent)
                r_k = 1 / self._model_setting.K[Tcurrent] * np.ones(self._model_setting.K[Tcurrent])
                gamma0 = 1
                c0 = 1
            else:
                self._model_setting.K[Tcurrent] = self._model_setting.K[Tcurrent - 1]
                if self._model_setting.K[Tcurrent] <= 4:
                    break
                self.Phi[Tcurrent] = np.random.rand(self._model_setting.K[Tcurrent - 1], self._model_setting.K[Tcurrent])
                self.Phi[Tcurrent] = self.Phi[Tcurrent] / np.maximum(realmin, np.sum(self.Phi[Tcurrent], 0))
                self.Theta[Tcurrent] = np.ones((self._model_setting.K[Tcurrent], self._model_setting.N)) / self._model_setting.K[Tcurrent]
                p_j = self._calculate_pj(c_j, Tcurrent)
                r_k = 1 / self._model_setting.K[Tcurrent] * np.ones(self._model_setting.K[Tcurrent])
                gamma0 = self._model_setting.K[Tcurrent] / self._model_setting.K[1]
                c0 = 1

            for iter in range(1, self._model_setting.Iteration[Tcurrent]):
                start_time = time.time()

                for t in range(Tcurrent + 1):
                    if t == 0:
                        dex111 = list(range(len(ZS)))
                        np.random.shuffle(dex111)
                        ZS = ZS[dex111]
                        DS = DS[dex111]
                        WS = WS[dex111]
                        if Tcurrent == 0:
                            shape = np.dot(r_k.reshape(-1, 1), np.ones((1, self._model_setting.N)))
                        else:
                            shape = np.dot(self.Phi[1], self.Theta[1])
                        beta1_sum = np.sum(beta1, 1)
                        # Modified from GNBP_mex_collapsed_deep.c in the GBN code,
                        # to support a full matrix of beta1
                        [self.local_params.Theta, temp, n_dot_k, ZS] = self._collapsed_gibbs_topic_assignment_mex(
                            self.local_params.Theta, self.global_params.Phi, n_dot_k, ZS, WS, DS, shape, beta1, beta1_sum)
                        if is_train:
                            self.global_params.Phi = temp
                        WSZS[t] = self.global_params.Phi.T
                        Xt_to_t1[t] = self.local_params.Theta
                        # Sample the variables related to sub-topics
                        beta1 = self.sample_beta(WSZS[t].T, embeddings, beta1)
                    else:
                        [Xt_to_t1[t], WSZS[t]] = self._sampler.multi_aug(Xt_to_t1[t-1], self.Phi[t], self.Theta[t])

                    if t > 0:
                        self.Phi[t] = self._sample_Phi(WSZS[t], beta0)
                        if np.count_nonzero(np.isnan(self.Phi[t])):
                            Warning('Phi Nan')
                            self.Phi[t][np.isnan(self.Phi[t])] = 0
                Xt = self._crt_sum_mex_matrix_v1(sparse.csc_matrix(Xt_to_t1[Tcurrent].T), r_k.reshape(1, -1).T).T
                r_k, gamma0, c0 = self._sample_rk(Xt, r_k, p_j[Tcurrent+1], gamma0, c0)

                if iter > 10:
                    if Tcurrent > 0:
                        p_j[1] = self._sampler.beta(np.sum(Xt_to_t1[0], 0)+a0, np.sum(self.Theta[1], 0)+b0)
                    else:
                        p_j[1] = self._sampler.beta(np.sum(Xt_to_t1[0], 0)+a0, np.sum(r_k)+b0)
                    p_j[1] = np.minimum(np.maximum(p_j[1], np.spacing(1)), 1-np.spacing(1))
                    c_j[1] = (1 - p_j[1]) / p_j[1]
                    for t in range(2, Tcurrent+2):
                        if t == Tcurrent+1:
                            c_j[t] = self._sampler.gamma(np.sum(r_k)*np.ones((1, self._model_setting.N))+e0) / (np.sum(self.Theta[t-1], 0)+f0)
                        else:
                            c_j[t] = self._sampler.gamma(np.sum(self.Theta[t], 0)+e0) / (np.sum(self.Theta[t-1], 0)+f0)
                    p_j_temp = self._calculate_pj(c_j, Tcurrent)
                    p_j[2:] = p_j_temp[2:]

                for t in range(Tcurrent, -1, -1):
                    if t == Tcurrent:
                        shape = r_k.reshape(-1, 1)
                    else:
                        shape = np.dot(self.Phi[t+1], self.Theta[t+1])
                    if t > 0:
                        self.Theta[t] = self._sampler.gamma(shape+Xt_to_t1[t]) * (1/(c_j[t+1] - np.log(np.maximum(1 - p_j[t], realmin))))
                        # (100, 12337/987)  (1, 12337)
                        if np.count_nonzero(np.isnan(self.Theta[t])):
                            Warning('Theta Nan')
                            self.Theta[t][np.isnan(self.Theta[t])] = 0

                end_time = time.time()
                stages = 'Training' if is_train else 'Testing'
                print(f'{stages} Stage: ',
                      f'Layer {Tcurrent:3d}, epoch {iter:3d} takes {end_time - start_time:.2f} seconds, topics {np.count_nonzero(Xt):3d}')

            for t in range(Tcurrent + 1):
                if t == 0:
                    self.Phi[t] = self._sample_Phi(WSZS[t], beta1.T, True)
                else:
                    self.Phi[t] = self._sample_Phi(WSZS[t], beta0, True)

            paraGlobal[Tcurrent]['Phi'] = self.Phi
            paraGlobal[Tcurrent]['r_k'] = r_k
            paraGlobal[Tcurrent]['gamma0'] = gamma0
            paraGlobal[Tcurrent]['c0'] = c0
            paraGlobal[Tcurrent]['K'] = self._model_setting.K[:Tcurrent]
            paraGlobal[Tcurrent]['beta0'] = beta0
            paraGlobal[Tcurrent]['beta_para'] = self.beta_para
            paraGlobal[Tcurrent]['p_j'] = p_j  # for theta
            paraGlobal[Tcurrent]['c_j'] = c_j
            paraGlobal[Tcurrent]['Xt_to_t1'] = Xt_to_t1
            paraGlobal[Tcurrent]['cjmedian'] = []
            for t in range(Tcurrent + 1):
                paraGlobal[Tcurrent]['cjmedian'].append(np.median(c_j[t]))

        return copy.deepcopy(self.local_params)


    def test(self, embeddings: np.ndarray, S: int, iter_all: list, data: np.ndarray):
        '''
        Inputs:
            embeddings     : [np.ndarray] V*D, word embedding of training words
            S              : [int] number of sub topics
            iter_all       : [np.ndarray] scalar, the iterations of gibbs sampling
            data           : [np.ndarray] V*N_train matrix, N_train bag-of-words vectors with a vocabulary length of V

        Outputs:
                local_params  : [Params] the local parameters of the probabilistic model

        '''
        local_params = self.train(embeddings, S, iter_all, data, is_train=False)

        return local_params


    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/WEDTM.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')


    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model;

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])


    def _init_beta(self, K, V, S, embeddings, beta):
        L = embeddings.shape[1]
        beta_para = [{}] * S
        for s in range(S):
            # variables for sub-topic s
            beta_para[s]['beta_s'] = beta/S * np.ones((K, V))
            beta_para[s]['alpha_k'] = 0.1 * np.ones((K, 1))
            beta_para[s]['W'] = 0.1 * np.ones((K, L))
            beta_para[s]['pi'] = np.dot(beta_para[s]['W'], embeddings.T)
            beta_para[s]['sigma'] = np.ones((K, L))
            beta_para[s]['c0'] = 1
            beta_para[s]['alpha0'] = 1

        beta1 = beta * np.ones((K, V))

        return beta1, beta_para


    def _calculate_pj(self, c_j, T):
        '''
        calculate p_j from layer 1 to T+1
        same as pfa
        '''
        p_j = [[]] * (T+2)
        N = len(c_j[1])
        p_j[0] = (1-np.exp(-1)) * np.ones((1, N))
        p_j[1] = 1/(1 + c_j[1])
        for t in range(2, T+2):
            temp = -np.log(np.maximum(1-p_j[t - 1], realmin))
            p_j[t] = temp / (temp + c_j[t])
            if np.count_nonzero(np.isnan(p_j[t])):
                Warning('pj Nan')
                p_j[t][np.isnan(p_j[t])] = np.spacing(1)
        return p_j


    def _collapsed_gibbs_topic_assignment_mex(self, ZSDS, ZSWS, n_dot_k, ZS, WS, DS, shape, eta, eta_sum):
        '''
        same as DirBN
        '''
        Ksize, Nsize = ZSDS.shape
        WordNum = WS.shape[0]
        prob_cumsum = np.zeros((Ksize, 1))

        for i in range(WordNum):
            v = WS[i]
            j = DS[i]
            k = ZS[i]
            if ZS[i] > -1:
                ZSDS[k, j] -= 1
                ZSWS[k, v] -= 1
                n_dot_k[k] -= 1
            cum_sum = 0
            for k in range(Ksize):
                cum_sum += (eta[k, v] + ZSWS[k, v]) / (eta_sum[k] + n_dot_k[k]) * (ZSDS[k, j] + shape[k, j])
                prob_cumsum[k] = cum_sum
            probrnd = np.random.rand() * cum_sum
            k = self._binary_search(probrnd, prob_cumsum, Ksize)
            ZS[i] = k
            ZSDS[k, j] += 1
            ZSWS[k, v] += 1
            n_dot_k[k] += 1

        return ZSDS, ZSWS, n_dot_k, ZS

    def _binary_search(self, probrnd, prob_cumsum, Ksize):
        if probrnd <= prob_cumsum[0]:
            return 0
        else:
            kstart = 1
            kend = Ksize - 1
            while 1:
                if kstart >= kend:
                    return kend
                else:
                    k = kstart + int((kend - kstart) / 2)
                    if (prob_cumsum[k - 1][0] > probrnd) & (prob_cumsum[k][0] > probrnd):
                        kend = k - 1
                    elif (prob_cumsum[k - 1][0] < probrnd) & (prob_cumsum[k][0] < probrnd):
                        kstart = k + 1
                    else:
                        return k
        return k


    def _sample_beta(self, n_topic_word, F, beta1):
        a0 = 0.01
        b0 = 0.01
        e0 = 1
        f0 = 1
        S = len(self.beta_para)
        L = F.shape[1]

        # The word count for each v and k in the first layer
        [K, V] = n_topic_word.shape
        n_sum = np.sum(n_topic_word, 1)

        ## Eq. (3)
        log_inv_q = -np.log(self._sampler.beta(np.sum(beta1, 1), np.maximum(n_sum, realmin)))
        log_log_inv_q = np.log(np.maximum(log_inv_q, realmin))

        # Active topics in the first layer
        active_k = (~np.isnan(log_inv_q)) & (~np.isinf(log_inv_q)) & (n_sum > 0) & (log_inv_q != 0)

        ## Eq. (4) and (6)
        h = np.zeros((K, V, S)).astype(int)
        for k in range(K):
            for v in range(V):
                for j in range(n_topic_word[k, v]):
                    if j == 0:
                        is_add_table = 1
                    else:
                        is_add_table = (np.random.rand() < beta1[k, v] / (beta1[k, v] + j + 1))
                    if is_add_table > 0:
                        p = np.zeros((S, 1))
                        for s in range(S):
                            p[s] = self.beta_para[s]['beta_s'][k, v]
                        sum_cum = np.cumsum(p)
                        temp = np.argwhere(sum_cum > np.random.rand() * sum_cum[-1])
                        if len(temp) > 0:
                            ss = temp[0]
                        else:
                            continue
                        h[k, v, ss] = h[k, v, ss] + 1
        beta1 = 0

        for s in range(S):
            ## For each sub-topic s
            alpha_k = self.beta_para[s]['alpha_k']
            pi_pg = self.beta_para[s]['pi']
            W = self.beta_para[s]['W']
            c0 = self.beta_para[s]['c0']
            alpha0 = self.beta_para[s]['alpha0']
            h_s = h[:, :, s]

            # Sample alpha_k for each sub-topic s with the hierarchical gamma
            h_st = np.zeros((K, V)).astype(int)
            # Eq. (11)
            h_st[h_s > 0] = 1
            for k in range(K):
                for v in range(V):
                    for j in range(h_s[k, v] - 1):
                        h_st[k, v] = h_st[k, v] + (np.random.rand() < alpha_k[k] / (alpha_k[k] + j + 1)).astype(int)

            # Eq. (10)
            h_st_dot = np.sum(h_st, 1)
            # Active topics in each sub-topic s
            local_active_k = h_st_dot > 0 & active_k
            l_a_K = sum(local_active_k)

            x = pi_pg + log_log_inv_q.reshape(-1, 1)
            dex = x < 0
            temp = np.zeros(x.shape)
            temp[dex] = np.log1p(np.exp(x[dex]))
            temp[~dex] = x[~dex]+np.log1p(np.exp(-x[~dex]))
            temp = np.sum(temp, 1)

            # Eq. (9)
            alpha_k = (self._sampler.gamma(alpha0 / l_a_K + h_st_dot) / (c0 + temp)).reshape(-1, 1)
            h_stt = np.zeros((K, 1))
            h_stt[h_st_dot > 0] = 1
            for k in range(K):
                for j in range(h_st_dot[k] - 1):
                    h_stt[k] = h_stt[k] + (np.random.rand() < (alpha0 / l_a_K) / (alpha0 / l_a_K + j + 1)).astype(int)

            temp2 = temp / (c0 + temp)
            # L17 in Figure 1 in the appendix
            alpha0 = self._sampler.gamma(a0 + np.sum(h_stt)) / (b0 - np.sum(np.log(1 - temp2[local_active_k])) / l_a_K)
            c0 = self._sampler.gamma(e0 + alpha0) / (f0 + np.sum(alpha_k[local_active_k]))

            ## Sample Polya-Gamma variables
            # Eq. (15)
            pi_pg_vec = pi_pg + log_log_inv_q.reshape(-1,1)
            pi_pg_vec = pi_pg_vec.reshape(K * V, 1)
            temp = h_s + alpha_k  # reshape(h_s + alpha_k, K*V,1)
            temp = temp.reshape(K * V, 1)
            omega_vec = self._polya_gam_rnd_gam(temp, pi_pg_vec, 2)
            omega_mat = omega_vec
            omega_mat = omega_mat.reshape(K, V)

            ## Sample sigma
            sigma_w = self._sampler.gamma(1e-2 + 0.5 * l_a_K) / (
                        1e-2 + np.sum(np.power(W[local_active_k, :], 2), 0) * 0.5)
            sigma_w = np.tile(sigma_w, (K, 1))

            ## Sample W
            # Eq. (14)
            for k in range(K):
                if local_active_k[k] > 0:
                    Hgam = F.T * omega_mat[k, :]
                    invSigmaW = np.diag(sigma_w[k, :]) + np.dot(Hgam, F)
                    MuW = np.dot(np.linalg.inv(invSigmaW), (
                        np.sum(F.T * (0.5 * h_s[k, :].reshape(1, -1) - 0.5 * alpha_k[k, :] - (log_log_inv_q[k]) * omega_mat[k, :]), 1)))
                    R = self._choll(invSigmaW)
                    W[k, :] = MuW + np.dot(np.linalg.inv(R), np.random.rand(L, 1)).flatten()
                else:
                    W[k, :] = 1e-10

            # Update pi, Eq. (8)
            pi_pg = np.dot(W, F.T)

            ## Sample beta for each sub-topic s
            # Eq. (7)
            beta_s = self._sampler.gamma(alpha_k + h_s) / (np.exp(-pi_pg) + log_inv_q.reshape(-1, 1))
            beta_s[local_active_k == 0, :] = 0.05 / S
            beta_s[(np.sum(np.isnan(beta_s), 1)) != 0, :] = 0.05 / S
            beta_s[(np.sum(np.isnan(beta_s) | np.isinf(beta_s), 1)) != 0, :] = 0.05 / S
            beta_s[(np.sum(beta_s, 1).astype(bool)), :] = 0.05 / S

            ## Update beta1
            beta1 = beta1 + beta_s

            ## Collect results
            # self.beta_para[s]['beta_s'] = beta_s
            self.beta_para[s]['pi'] = pi_pg
            self.beta_para[s]['W'] = W
            self.beta_para[s]['alpha_k'] = alpha_k
            self.beta_para[s]['sigma'] = sigma_w
            self.beta_para[s]['h_s'] = sparse.csc_matrix(h_s)
            self.beta_para[s]['c0'] = c0
            self.beta_para[s]['alpha0'] = alpha0

        return beta1


    def _polya_gam_rnd_gam(self, a, c, KK, IsBiased=None):
        '''
        Generating Polya-Gamma random varaibles using approximation method
        '''
        IsBiased = False
        x = 1 / 2 / math.pi ** 2 * np.sum(self._sampler.gamma(np.dot(a, np.ones((1, KK))), 1) /
                                          (np.power((np.array([i for i in range(KK)]) + 0.5), 2) + np.power(c, 2) / 4 / math.pi ** 2), 1)

        if ~IsBiased:
            temp = abs(c / 2)
            temp[temp <= 0] = realmin
            xmeanfull = (np.tanh(temp) / (temp) / 4)
            xmeantruncate = 1 / 2 / math.pi ** 2 * np.sum(
                1 / (np.power((np.array([i for i in range(KK)]) + 0.5), 2) + np.power(c, 2) / 4 / math.pi ** 2), 1)
            x = x * xmeanfull.flatten() / (xmeantruncate)

        return x

    def _choll(self, A):
        # same as dpfa
        P = A.copy()
        q = np.linalg.cholesky(P)
        q = q.T

        return q

    def _crt_sum_mex_matrix_v1(self, X, r):
        # same as DirBN sample_theta
        k, n = np.shape(X)
        if len(r) == 1:
            r = r[0]
        lsum = np.zeros(n).astype(int)
        maxx = 0
        Xsparse = sparse.csc_matrix(X)
        pr = Xsparse.data
        ir = Xsparse.indices
        jc = Xsparse.indptr

        for j in range(n):
            starting_row_index = jc[j]
            stopping_row_index = jc[j+1]
            if starting_row_index == stopping_row_index:
                continue
            else:
                for current_row_index in range(starting_row_index, stopping_row_index):
                    maxx = int(max(maxx, pr[current_row_index]))
                prob = np.zeros(maxx)
                for i in range(maxx):
                    prob[i] = r[j] / (r[j] + i)

                lsum[j] = 0
                for current_row_index in range(starting_row_index, stopping_row_index):
                    for i in range(int(pr[current_row_index])):
                        if np.random.rand() <= prob[i]:
                            lsum[j] += 1

        return lsum

    def _sample_rk(self,XTplusOne_sum=None, r_k=None, p_jTplusOne=None, gamma0=None, c0=None, IsNoSample=None, e0=None,
                  f0=None, a0=None, b0=None):
        '''
        get theta_para.r_k, theta_para.gamma0, theta_para.c0
        '''
        if len(p_jTplusOne) == 1:
            p_jTplusOne = p_jTplusOne[0]
        IsNoSample = False
        e0 = 1
        f0 = 1
        a0 = 0.01
        b0 = 0.01
        if (len(XTplusOne_sum.shape) > 1):
            if (XTplusOne_sum.shape[1] > 1):
                # XTplusOne_sum = np.full(np.sum(XTplusOne_sum, 1))
                print('unexcepted error')

        KT = len(r_k)
        if ~IsNoSample:
            c0 = self._sampler.gamma(e0 + gamma0) / (f0 + np.sum(r_k))
            temp = 1 - p_jTplusOne
            temp[temp <= 0] = 2.2251e-308
            sumlogpi = np.sum(np.log(temp))
            p_prime = - sumlogpi / (c0 - sumlogpi)
            # L_k = full(sum(XTplusOne,2));
            # XTplusOne_sum
            gamma0 = self._sampler.gamma(a0 + self.CRT_sum_mex_v1(XTplusOne_sum, (gamma0 / KT))) / (b0 - np.log(max(1 - p_prime, 2.2251e-308)))
            r_k = self._sampler.gamma(gamma0 / KT + XTplusOne_sum) / (c0 - sumlogpi)

        else:
            print('unexcepted error')

        return r_k, gamma0, c0

    def _crt_sum_mex_v1(self, x, r):
        # same to dpfa-CRT
        xx = np.unique(x)
        jj = np.array([np.argwhere(xx == t) for t in x.flatten()]).flatten()
        L = np.zeros(len(x))
        Lsum = 0
        if not x is None:
            for i in range(len(xx)):
                y = int(xx[i])
                if y > 0:
                    L[jj == i] = np.sum(np.random.rand(np.count_nonzero(jj == i), y) <= (r / (r + np.array([t for t in range(y)]))), 1)[0]
            Lsum = int(sum(L))

        return Lsum

    def _sample_Phi(self, WSZS, Eta, IsNoSample=False):
        if ~IsNoSample:
            Phi = self._sampler.gamma(Eta + WSZS)
            temp = np.sum(Phi, 0)
            tempdex = temp > 0
            Phi[:, tempdex] = Phi[:, tempdex] / temp[tempdex]
            Phi[:, ~tempdex] = 0
            if np.count_nonzero(np.isnan(Phi)):
                Warning('Phi Nan')
                tempdex = temp > 0
                Phi[:, ~tempdex] = 0
        else:
            Phi = Eta + WSZS
            temp = np.sum(Phi, 0)
            Phi = Phi / temp
            if np.count_nonzero(np.isnan(Phi)):
                Warning('Phi Nan')
                tempdex = temp > 0
                Phi[:, ~tempdex] = 0
        return Phi

    # collapsed_gibbs_topic_assignment_mex from dirbn.collapsed_gibbs_topic_assignment_mex
    # TrimTcurrent_WEDTM -to-do but not necessary. Prune the inactive factors of the current top hidden layer
    # code before 'iter>10' is similar to sample_theta from DirBN
    # sample_beta func is similar to sample_DirBN_beta, sample_DirBN_counts from DirBN, but been polished
    # some code of sample_beta is similar to dpfa - train 3..
    # PolyaGamRnd_Gam and PolyaGamRndTruncated(dpfa), similar but not same
    # choll is from dpfa






from ._basic_model import Params, Basic_Model
from ._lda import LDA
from ._pfa import PFA
from ._pgbn import PGBN
from ._cpfa import CPFA
from ._cpgbn import CPGBN
from ._pgds import PGDS
from ._dpgds import DPGDS
from ._wedtm import WEDTM
from ._dirbn import DirBN
from ._dpfa import DPFA
from ._gpgbn import GPGBN
from ._mpgbn import MPGBN


'''
example of models and GPU distribution sampler.
To run models demo, you need to download data files 'mnist_gray.mat'&'TREC.pkl' and put it under pydpm.example.data
data url: https://1drv.ms/u/s!AlkDawhaUUBWtHRWuNESEdOsDz7V?e=LQlGLW
'''



"""
===========================================
WEDTM Demo
Inter and Intra Topic Structure Learning with Word Embeddings
He Zhao, Lan Du, Wray Buntine, Mingyaun Zhou
Published in International Council for Machinery Lubrication 2018

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import numpy as np
import scipy.io as sio

from pydpm._metric import ACC
from pydpm._model import WEDTM

# load data
dataset = sio.loadmat('./data/WS.mat')
train_data = np.asarray(dataset['doc'].todense()[:, dataset['train_idx'][0]-1])[:, ::10].astype(int)
test_data = np.asarray(dataset['doc'].todense()[:, dataset['test_idx'][0]-1])[:, ::5].astype(int)
train_label = dataset['labels'][dataset['train_idx'][0]-1][::10, :]
test_label = dataset['labels'][dataset['test_idx'][0]-1][::5, :]

# params of model
T = 3  # vertical layers
S = 3  # sub topics
K = [100] * T  # topics in each layers

# create the model and deploy it on gpu or cpu
model = WEDTM(K, 'gpu')
model.initial(dataset['doc'])  # use the shape of train_data to initialize the params of model
train_local_params = model.train(dataset['embeddings'], S, 300, train_data)
train_local_params = model.test(dataset['embeddings'], S, 300, train_data)
test_local_params = model.test(dataset['embeddings'], S, 300, test_data)

# evaluate the model with classification accuracy
# the demo accuracy can achieve
results = ACC(train_local_params.Theta, test_local_params.Theta, train_label, test_label, 'SVM')

# save the model after training
model.save()
# model.load('./save_models/WEDTM.npy')





from ._accuracy import ACC
from ._reconstruction import Poisson_Likelihood, Reconstruct_Error
from ._cluster_acc import Cluster_ACC
from ._topic_coherence import Topic_Coherence
from ._perplexity import Perplexity
from ._normalized_mutual_information import NMI



import os
import numpy as np
import ctypes

class distribution_sampler_cpu(object):

    def __init__(self):
        """
        The basic class for sampling distribution on cpu
        """
        super(distribution_sampler_cpu, self).__init__()

        # sampler for basic distributions
        setattr(self, 'standard_normal', np.random.standard_normal)
        setattr(self, 'normal', np.random.normal)
        setattr(self, 'standard_gamma', np.random.standard_gamma)
        setattr(self, 'gamma', np.random.gamma)
        setattr(self, 'standard_cauchy', np.random.standard_cauchy)
        # setattr(self, 'cauchy', np.random.cauchy) # numpy doesnot has this distribution
        setattr(self, 'chisquare', np.random.chisquare)
        setattr(self, 'beta', np.random.beta)
        # setattr(self, 'crt', np.random.crt) # numpy doesnot has this distribution
        setattr(self, 'dirichlet', np.random.dirichlet)
        setattr(self, 'poisson', np.random.poisson)
        setattr(self, 'weibull', np.random.weibull)
        setattr(self, 'negative_binomial', np.random.negative_binomial)
        setattr(self, 'lognormal', np.random.lognormal)
        setattr(self, 'binomial', np.random.binomial)
        setattr(self, 'multinomial', np.random.multinomial)
        setattr(self, 'laplace', np.random.laplace)
        setattr(self, 'logistic', np.random.logistic)
        setattr(self, 'exponential', np.random.exponential)
        setattr(self, 'standard_exponential', np.random.standard_exponential)
        setattr(self, 'noncentral_chisquare', np.random.noncentral_chisquare)
        setattr(self, 'zipf', np.random.zipf)
        setattr(self, 'triangular', np.random.triangular)
        setattr(self, 'noncentral_f', np.random.noncentral_f)
        setattr(self, '_f', np.random.f)
        # setattr(self, 't', np.random.t) # numpy doesnot has this distribution
        setattr(self, 'geometric', np.random.geometric)
        setattr(self, 'hypergeometric', np.random.hypergeometric)
        setattr(self, 'gumbel', np.random.gumbel)
        setattr(self, 'pareto', np.random.pareto)
        setattr(self, 'power', np.random.power)
        setattr(self, 'rayleigh', np.random.rayleigh)



"""
===========================================
per-held-word perplexity
===========================================

"""

import numpy as np
from .._utils import *

class Perplexity(object):
	def __init__(self, x, x_hat):
		'''
		Inputs:
			x: [float] np.ndarray, V*N_test matrix, the observations for test_data
			x_hat: [float] np.ndarray, V*N_reconstruct matrix
		Outputs:
			PPL: [float], the perplexity score
		'''

		self.x = x
		self.x_hat = x_hat

		self._get()

		print(f'The PPL is: {self._PPL:.4f}')

	def _get(self):

		self.x_hat = self.x_hat / (self.x_hat.sum(0) + realmin)
		ppl = -1.0 * self.x * np.log(self.x_hat.T + realmin) / self.x.sum()
		ppl = ppl.sum().exp()

		self._PPL = ppl



"""
===========================================
Model Sampler implemented on CPU
===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu
# License: BSD-3-Clause

import numpy as np
import numpy.ctypeslib as npct
import ctypes  
from ctypes import *
import os

class model_sampler_cpu(object):

    def __init__(self, system_type='Windows', seed=0):
        """
        The basic class for sampling distribution on cpu
        """
        super(model_sampler_cpu, self).__init__()

        self.system_type = system_type
        self.seed = seed

        array_2d_double = npct.ndpointer(dtype=np.double, ndim=2, flags='C')
        array_1d_double = npct.ndpointer(dtype=np.double, ndim=1, flags='C')
        array_int = npct.ndpointer(dtype=np.int32, ndim=0, flags='C')
        ll = ctypes.cdll.LoadLibrary

        if system_type == "Windows":
            self.Crt_lib = ll(os.path.dirname(__file__) + "\_compact\crt_cpu.dll")
            self.Multi_lib = ll(os.path.dirname(__file__) + "\_compact\multi_aug_cpu.dll")
            self.Crt_Multi_lib = ll(os.path.dirname(__file__) + "\_compact\crt_multi_aug_cpu.dll")
        else:
            self.Crt_lib = ll(os.path.dirname(__file__) + "/_compact/crt_cpu.so")
            self.Multi_lib = ll(os.path.dirname(__file__) + "/_compact/multi_aug_cpu.so")
            self.Crt_Multi_lib = ll(os.path.dirname(__file__) + "/_compact/crt_multi_aug_cpu.so")


        self.Multi_lib.Multi_Sample.restype = None
        self.Multi_lib.Multi_Sample.argtypes = [array_2d_double, array_2d_double, array_2d_double, array_2d_double,
                                           array_2d_double, c_int, c_int, c_int]

        self.Crt_Multi_lib.Crt_Multi_Sample.restype = None
        self.Crt_Multi_lib.Crt_Multi_Sample.argtypes = [array_2d_double, array_2d_double, array_2d_double, array_2d_double,
                                                   array_2d_double, c_int, c_int, c_int]

        self.Crt_lib.Crt_Sample.restype = None
        self.Crt_lib.Crt_Sample.argtypes = [array_2d_double, array_2d_double, array_2d_double, c_int, c_int]

    def multi_aug(self, X_t, Phi_t, Theta_t):

        X_t = np.array(X_t, order='C').astype('double')
        Phi_t = np.array(Phi_t, order='C').astype('double')
        Theta_t = np.array(Theta_t, order='C').astype('double')

        V = X_t.shape[0]
        J = X_t.shape[1]
        K = Theta_t.shape[0]
        Xt_to_t1_t = np.zeros([K, J], order='C').astype('double')
        WSZS_t = np.zeros([V, K], order='C').astype('double')
        self.Multi_lib.Multi_Sample(X_t, Phi_t, Theta_t, WSZS_t, Xt_to_t1_t, V, K, J)

        return Xt_to_t1_t, WSZS_t

    def crt(self, Xt_to_t1_t, p):

        Xt_to_t1_t = np.array(Xt_to_t1_t, order='C')
        p = np.array(p, order='C')

        K_t = Xt_to_t1_t.shape[0]
        J = Xt_to_t1_t.shape[1]
        X_t1 = np.zeros([K_t, J], order='C').astype('double')

        self.Crt_lib.Crt_Sample(Xt_to_t1_t, p, X_t1, K_t, J)

        return X_t1

    def crt_multi_aug(self, Xt_to_t1_t, Phi_t1, Theta_t1):

        Xt_to_t1_t = np.array(Xt_to_t1_t, order='C').astype('double')
        Phi_t1 = np.array(Phi_t1, order='C').astype('double')
        Theta_t1 = np.array(Theta_t1, order='C').astype('double')

        K_t = Xt_to_t1_t.shape[0]
        J = Xt_to_t1_t.shape[1]
        K_t1 = Theta_t1.shape[0]
        Xt_to_t1_t1 = np.zeros([K_t1, J], order='C').astype('double')
        WSZS_t1 = np.zeros([K_t, K_t1], order='C').astype('double')

        self.Crt_Multi_lib.Crt_Multi_Sample(Xt_to_t1_t, Phi_t1, Theta_t1, WSZS_t1, Xt_to_t1_t1, K_t, K_t1, J)

        return Xt_to_t1_t1, WSZS_t1






import numpy as np
import copy

class NMI(object):
    def __init__(self, A, B):
        '''
        Inputs:
            A: [int], ground_truth, shape:(n_sample,)
            B: [int], pred_label, shape:(n_sample,)

        Outputs:
            NMI: [float], Normalized Mutual information of A and B

        '''
        self.A = copy.deepcopy(A)
        self.B = copy.deepcopy(B)

        self._get()

        print(f'The NMI is: {self._NMI:.4f}')

    def _get(self):
        n_gnd = self.A.shape[0]
        n_label = self.B.shape[0]
        # assert n_gnd == n_label

        LabelA = np.unique(self.A)
        nClassA = len(LabelA)
        LabelB = np.unique(self.B)
        nClassB = len(LabelB)

        if nClassB < nClassA:
            self.A = np.concatenate((self.A, LabelA))
            self.B = np.concatenate((self.B, LabelA))
        else:
            self.A = np.concatenate((self.A, LabelB))
            self.B = np.concatenate((self.B, LabelB))

        G = np.zeros([nClassA, nClassA])
        for i in range(nClassA):
            for j in range(nClassA):
                G[i, j] = np.sum((self.A == LabelA[i]) * (self.B == LabelA[j]))

        sum_G = np.sum(G)
        PA = np.sum(G, axis=1)
        PA = PA/sum_G
        PB = np.sum(G, axis=0)
        PB = PB/sum_G
        PAB = G/sum_G

        if np.sum((PA == 0)) > 0 or np.sum((PB == 0)):
            print('error ! Smooth fail !')
            self._NMI = np.nan
        else:
            HA = np.sum(-PA * np.log2(PA))
            HB = np.sum(-PB * np.log2(PB))
            PPP = PAB / np.tile(PB, (nClassA, 1)) / np.tile(PA.reshape(-1, 1), (1, nClassA))
            PPP[np.where(abs(PPP) < 1E-12)] = 1  # avoid 'log 0'
            MI = np.sum(PAB * np.log2(PPP))
            NMI = MI / np.max((HA, HB))
            # optional
            # NMI = 2.0 * MI / (HA + HB)
            # NMI = MI / np.sqrt(HA * HB)
            self._NMI = NMI




"""
===========================================
Poisson Factor Analysis DirBN(Dirichlet belief networks) Demo
Dirichlet belief networks for topic structure learning
He Zhao, Lan Du, Wray Buntine, Mingyuan Zhou
Publihsed in Conference and Workshop on Neural Information Processing Systems 2018

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause


import os
import copy
import time
import numpy as np

from ._basic_model import Basic_Model
from .._sampler import Basic_Sampler
from .._utils import *

from scipy import sparse


class DirBN(Basic_Model):
    def __init__(self, K: list, device='gpu'):
        """
        The basic model for DirBN
        Inputs:
            K      : [list] number of topics of 2 layers in DirBN;
            device : [str] 'cpu' or 'gpu';

        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(DirBN, self).__init__()
        setattr(self, '_model_name', 'DirBN')

        self._model_setting.K = K
        self._model_setting.device = device

        assert self._model_setting.device in ['cpu', 'gpu'], 'Device Type Error: the device should be ''cpu'' or ''gpu'''

        self._sampler = Basic_Sampler(self._model_setting.device)


    def initial(self, data: np.ndarray):
        '''
        Inintial the parameters of DirBN with the input documents
        Inputs:
            data : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary length of V

        Attributes:
            @public:
                global_params.Phi  : [np.ndarray] V*K matrix, K topics with a vocabulary length of V
                local_params.Theta : [np.ndarray] N*K matrix, the topic propotions of N documents

            @private:
                _model_setting.V        : [int] scalar, the length of the vocabulary
                _hyper_params.Phi_eta   : [float] scalar, the parameter in the prior of Phi

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.V = data.shape[0]

        self.global_params.Phi = np.zeros((self._model_setting.K[0], self._model_setting.V)).astype(int)  # frequency, not distribution

        self._hyper_params.Phi_eta = 0.05

    def train(self, iter_all: int, data: np.ndarray, is_train: bool = True):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V
            is_train   : [bool] True or False, whether to update the global params in the probabilistic model

        Attributes:
            @public:
                local_params.Theta : [np.ndarray] N_train*K matrix, the topic propotions of N_train documents

            @private:
                _model_setting.N         : [int] scalar, the number of the documents in the corpus
                _model_setting.Iteration : [int] scalar, the iterations of gibbs sampling
                _hyper_params.DirBN_para : [dict] scalar, the hyper params of DirBN
                _hyper_params.theta_para : [dict] scalar, the hyper params theta of model

        Outputs:
                local_params  : [Params] the local parameters of the probabilistic model

        '''
        assert type(data) is np.ndarray, 'Data type error: the input data should be a 2-D np.ndarray'
        self._model_setting.N = data.shape[1]
        self._model_setting.Iteration = iter_all

        # initial local parameters
        self.local_params.Theta = np.zeros((self._model_setting.K[0], self._model_setting.N)).astype(int)

        # initial _hyper_params
        self._hyper_params.DirBN_para = self._init_DirBN(self._model_setting.K, self._model_setting.V, self._hyper_params.Phi_eta)
        self._hyper_params.theta_para = self._init_theta(self._model_setting.K[0], self._model_setting.N)

        self.data = dict()
        self.data['train_ws'] = []
        self.data['train_ds'] = []
        for n in range(self._model_setting.N):
            for v in range(self._model_setting.V):
                self.data['train_ws'].extend([v] * int(data[v, n]))
                self.data['train_ds'].extend([n] * int(data[v, n]))
        self.data['train_ws'] = np.array(self.data['train_ws'])
        self.data['train_ds'] = np.array(self.data['train_ds'])

        self.zs = np.random.randint(self._model_setting.K[0], size=int(np.sum(data)))  # theme of words
        # self.zs_ds, self.local_params.Theta, theme frequency of docs
        if is_train:
            self.global_params.Phi = np.zeros((self._model_setting.K[0], self._model_setting.V)).astype(int)  # phi: distribution of words
        for i in range(len(self.zs)):
            self.local_params.Theta[self.zs[i], self.data['train_ds'][i]] += 1
            if is_train:  # when train
                self.global_params.Phi[self.zs[i], self.data['train_ws'][i]] += 1  # word frequency, not distribution

        self.n_dot_k = np.sum(self.local_params.Theta, 1)  # K, frequency of themes

        for iter in range(self._model_setting.Iteration):
            start_time = time.time()

            # sample topic assignments by the collapsed Gibbs sampling
            self.local_params.Theta, temp, self.n_dot_k, self.zs = self._collapsed_gibbs_topic_assignment_mex(self.local_params.Theta, copy.deepcopy(self.global_params.Phi), self.n_dot_k, self.zs, self.data['train_ws'],
                                                                             self.data['train_ds'],
                                                                             np.tile(self._hyper_params.theta_para['r_k'], (1, self._model_setting.N)),
                                                                             self._hyper_params.DirBN_para[0]['psi'],
                                                                             np.sum(self._hyper_params.DirBN_para[0]['psi'], 1))
            if is_train:
                self.global_params.Phi = temp

            self._sample_DirBN(self.global_params.Phi)  # get self._hyper_params.DirBN_para
            self._sample_theta(self.local_params.Theta)  # get self._hyper_params.theta_para

            end_time = time.time()
            stages = 'Training' if is_train else 'Testing'
            print(f'{stages} Stage: ',
                  f'epoch {iter:3d} takes {end_time - start_time:.2f} seconds')

        return copy.deepcopy(self.local_params)


    def test(self, iter_all: int, data: np.ndarray):
        '''
        Inputs:
            iter_all   : [int] scalar, the iterations of gibbs sampling
            data       : [np.ndarray] V*N matrix, N bag-of-words vectors with a vocabulary of length V

        Outputs:
            local_params  : [Params] the local parameters of the probabilistic model

        '''
        local_params = self.train(iter_all, data, is_train=False)

        return local_params


    def save(self, model_path: str = './save_models'):
        '''
        Save the model to the specified directory.
        Inputs:
            model_path : [str] the directory path to save the model, default './save_models/DirBN.npy'
        '''
        # create the directory path
        if not os.path.isdir(model_path):
            os.mkdir(model_path)

        # save the model
        model = {}
        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in dir(self):
                model[params] = getattr(self, params)

        np.save(model_path + '/' + self._model_name + '.npy', model)
        print('model have been saved by ' + model_path + '/' + self._model_name + '.npy')


    def load(self, model_path: str):
        '''
        Load the model parameters from the specified directory
        Inputs:
            model_path : [str] the directory path to load the model;

        '''
        assert os.path.exists(model_path), 'Path Error: can not find the path to load the model'
        model = np.load(model_path, allow_pickle=True).item()

        for params in ['global_params', 'local_params', '_model_setting', '_hyper_params']:
            if params in model:
                setattr(self, params, model[params])

    def _init_DirBN(self, ks, V, eta):
        '''
        initial DirBN params
        '''
        T = len(ks)
        DirBN_para = []
        for t in range(T):
            DirBN_para.append({'None': None})
        '''
        [0]
                 psi: [100×13370 double]
        n_topic_word: [100×13370 double]
                 phi: [100×13370 double]
        [1]
                  psi: [100×13370 double]
                  phi: [100×13370 double]
                 beta: [100×100 double]
          beta_gammak: [100×1 double]
               beta_c: 1.0298
          beta_gamma0: 16.8023
              beta_c0: 0.3564
         n_topic_word: [100×13370 double]
        n_topic_topic: [100×100 double]'''
        for t in range(T):
            DirBN_para[t]['psi'] = eta * np.ones((ks[t], V))
            if t > 0:
                if t < T - 1:
                    DirBN_para[t]['phi'] = self._sampler.gamma(DirBN_para[t]['psi'])
                else:
                    DirBN_para[t]['phi'] = self._sampler.gamma(DirBN_para[t]['psi'] * np.ones((ks[t], V)))
                DirBN_para[t]['phi'] = DirBN_para[t]['phi'] / DirBN_para[t]['phi'].sum(axis=1, keepdims=1)
                DirBN_para[t]['beta'] = 0.5 * np.ones((ks[t], ks[t - 1]))
                DirBN_para[t]['beta_gammak'] = 0.1 * np.ones((ks[t], 1))
                DirBN_para[t]['beta_c'] = 0.1
                DirBN_para[t]['beta_gamma0'] = 1.0
                DirBN_para[t]['beta_c0'] = 1.0
        return DirBN_para

    def _init_theta(self, K = None, N = None):
        theta_para = {'gamma0': 1, 'c0': 1, 'r_k': 1 / K*np.ones((K, 1)), 'p_j': (1 - np.exp(- 1)) * np.ones((1, N))}
        return theta_para

    def _collapsed_gibbs_topic_assignment_mex(self, ZSDS, ZSWS, n_dot_k, ZS, WS, DS, shape, eta, eta_sum):
        Ksize, Nsize = ZSDS.shape
        WordNum = WS.shape[0]
        prob_cumsum = np.zeros(Ksize)

        # collapsed_gibbs_topic_assignment
        for i in range(WordNum):
            v = WS[i]  # index of word
            j = DS[i]  # index of doc
            k = ZS[i]  # index of theme
            if ZS[i] > -1:
                ZSDS[k, j] -= 1
                ZSWS[k, v] -= 1
                n_dot_k[k] -= 1
            cum_sum = 0
            for k in range(Ksize):  # for each themes
                cum_sum += (eta[k, v] + ZSWS[k, v]) / (eta_sum[k] + n_dot_k[k]) * (ZSDS[k, j] + shape[k, j])
                prob_cumsum[k] = cum_sum
            probrnd = np.random.rand() * cum_sum
            k = self._binary_search(probrnd, prob_cumsum, Ksize)
            ZS[i] = k
            ZSDS[k, j] += 1
            ZSWS[k, v] += 1
            n_dot_k[k] += 1

        return ZSDS, ZSWS, n_dot_k, ZS

    def _binary_search(self, probrnd, prob_cumsum, Ksize):
        if probrnd <= prob_cumsum[0]:
            return 0
        else:
            kstart = 1
            kend = Ksize - 1
            while 1:
                if kstart >= kend:
                    return kend
                else:
                    k = kstart + int((kend - kstart) / 2)
                    if (prob_cumsum[k-1] > probrnd) & (prob_cumsum[k] > probrnd):
                        kend = k - 1
                    elif (prob_cumsum[k-1] < probrnd) & (prob_cumsum[k] < probrnd):
                        kstart = k + 1
                    else:
                        return k

        return k

    def _sample_DirBN(self, n_topic_word1 = None):
        T = len(self._hyper_params.DirBN_para)
        T_current = T
        self._hyper_params.DirBN_para[0]['n_topic_word'] = n_topic_word1
        if T > 1:
            # propagate the latent counts from the bottom up
            for t in range(T_current - 1):
                self._sample_DirBN_counts(t)
            # update the latent variables from the top down
            for t in range(T_current-1, -1, -1):
                # update psi
                if t < T-1:
                    self._hyper_params.DirBN_para[t]['psi'] = np.dot(self._hyper_params.DirBN_para[t+1]['beta'].T, self._hyper_params.DirBN_para[t+1]['phi'])
                else:
                    psi = self._sample_DirBN_eta(self._hyper_params.DirBN_para[T-1]['psi'][0, 0], self._hyper_params.DirBN_para[T-1]['n_topic_word'])
                    self._hyper_params.DirBN_para[T-1]['psi'] = np.ones_like(self._hyper_params.DirBN_para[T-1]['n_topic_word']) * psi
                # update beta
                if t > 0:
                    self._sample_DirBN_beta(t)
                # update phi
                phi = self._sampler.gamma(np.abs(self._hyper_params.DirBN_para[t]['psi'] + self._hyper_params.DirBN_para[t]['n_topic_word'])) + 2.2204e-16
                phi = phi / phi.sum(axis=1, keepdims=1)
                self._hyper_params.DirBN_para[t]['phi'] = phi
        else:
            psi = self._sample_DirBN_eta(self._hyper_params.DirBN_para[T-1]['psi'][0, 0], self._hyper_params.DirBN_para[T-1]['n_topic_word'])
            self._hyper_params.DirBN_para[T-1]['psi'] = np.ones_like(n_topic_word1) * psi
            phi = self._sampler.gamma(self._hyper_params.DirBN_para[T-1]['psi'] + self._hyper_params.DirBN_para[T-1]['n_topic_word']) + 2.2204e-16
            phi = phi / (np.sum(phi, 1).reshape(-1, 1))
            self._hyper_params.DirBN_para[T-1]['phi'] = phi

    def _sample_DirBN_eta(self, eta=None, n=None):
        mu_0 = 0.1
        nu_0 = 10.0
        K, V = n.shape
        # if (np.sum(V*eta<0)+np.sum(np.sum(n, 1)<0))>0:
        #     raise('L341, negative value')
        #     log_q = -np.log(np.random.beta(np.maximum(V*eta, realmin), np.maximum(np.sum(n, 1), realmin)))
        # else:
        #     log_q = -np.log(np.random.beta(V*eta, np.sum(n, 1)))
        log_q = -np.log(self._sampler.beta(np.maximum(V * eta, realmin), np.maximum(np.sum(n, 1), realmin)))
        t = np.zeros((K, V))
        t[n > 0] = 1
        for k in range(K):
            for v in range(V):
                for j in range(1, int(n[k, v])):
                    t[k, v] = t[k, v] + (np.random.rand() < eta / (eta + j))

        eta = self._sampler.gamma(mu_0 + np.sum(t)) / (nu_0 + V * np.sum(log_q))

        return eta

    def _sample_DirBN_beta(self, t=None):
        a0 = 0.01
        b0 = 0.01
        e0 = 1
        f0 = 1
        beta_gamma0 = self._hyper_params.DirBN_para[t]['beta_gamma0']
        beta_gammak = self._hyper_params.DirBN_para[t]['beta_gammak']
        beta_c0 = self._hyper_params.DirBN_para[t]['beta_c0']
        beta_c = self._hyper_params.DirBN_para[t]['beta_c']

        if (np.sum(np.sum(self._hyper_params.DirBN_para[t - 1]['psi'], 1) < 0) + np.sum(np.sum(self._hyper_params.DirBN_para[t-1]['n_topic_word'], 1) < 0)) > 0:
            Warning('negative value')
        w_log_inv_q = -np.log(self._sampler.beta(np.maximum(np.sum(self._hyper_params.DirBN_para[t - 1]['psi'], 1), realmin),
                                              np.maximum(np.sum(self._hyper_params.DirBN_para[t-1]['n_topic_word'], 1), realmin)))
        w_t_k2_k1 = self._hyper_params.DirBN_para[t]['n_topic_topic']
        K2, K1 = w_t_k2_k1.shape
        w_tt = np.zeros([K2, K1])
        w_tt[w_t_k2_k1 > 0] = 1

        for k2 in range(K2):
            for k1 in range(K1):
                for j in range(1, int(w_t_k2_k1[k2, k1])):
                    w_tt[k2, k1] = w_tt[k2, k1] + (np.random.rand() < beta_gammak[k2] / (beta_gammak[k2] + j))

        w_tt_k2_dot = np.sum(w_tt, 1)
        active_k1 = (~(np.isnan(w_log_inv_q))) & (~(np.isinf(w_log_inv_q))) & (w_log_inv_q != 0)
        a_K1 = np.sum(active_k1)
        temp = np.log(1 + w_log_inv_q / beta_c)
        temp = np.sum(temp[active_k1])
        beta_gammak = self._sampler.gamma(beta_gamma0 / K2 + w_tt_k2_dot) / (beta_c0 + temp)
        w_tt_k2_dot_t = np.zeros(K2)
        w_tt_k2_dot_t[w_tt_k2_dot > 0] = 1

        for k2 in range(K2):
            for j in range(1, int(w_tt_k2_dot[k2])):
                w_tt_k2_dot_t[k2] = w_tt_k2_dot_t[k2] + (np.random.rand() < (beta_gamma0 / K2) / (beta_gamma0 / K2 + j))

        beta_gamma0 = self._sampler.gamma(a0 + np.sum(w_tt_k2_dot_t)) / (b0 + np.log(1 + temp / beta_c0))
        beta_c0 = self._sampler.gamma(e0 + beta_gamma0) / (f0 + np.sum(beta_gammak))
        beta_c = self._sampler.gamma(1.0 + a_K1 * np.sum(beta_gammak)) / (1.0 + np.sum(self._hyper_params.DirBN_para[t]['beta']))
        DirBN_beta = self._sampler.gamma(w_t_k2_k1 + beta_gammak.reshape(-1, 1)) / (beta_c + np.tile(w_log_inv_q.reshape(-1 ,1), (1, K2)).T)

        self._hyper_params.DirBN_para[t]['beta_gammak'] = beta_gammak
        self._hyper_params.DirBN_para[t]['beta'][:, active_k1] = DirBN_beta[:, active_k1]
        self._hyper_params.DirBN_para[t]['beta_c'] = beta_c
        self._hyper_params.DirBN_para[t]['beta_gamma0'] = beta_gamma0
        self._hyper_params.DirBN_para[t]['beta_c0'] = beta_c0

    def _sample_DirBN_counts(self, t=None):
        n_topic_word = self._hyper_params.DirBN_para[t]['n_topic_word']
        phi = self._hyper_params.DirBN_para[t + 1]['phi']
        DirBN_beta = self._hyper_params.DirBN_para[t + 1]['beta']
        DirBN_psi = self._hyper_params.DirBN_para[t]['psi']
        K1, V = n_topic_word.shape
        K2 = phi.shape[0]
        w_t_k2_k1 = np.zeros((K2, K1))
        w_t_k2_v = np.zeros((K2, V))

        for k1 in range(K1):
            for v in range(V):
                for j in range(int(n_topic_word[k1, v])):
                    if j == 0:
                        if_t = 1
                    else:
                        if_t = np.random.rand() < DirBN_psi[k1, v] / (DirBN_psi[k1, v] + j+1)
                    if if_t > 0:
                        p = phi[:, v] * DirBN_beta[:, k1]
                        sum_cum = np.cumsum(p)
                        k2 = np.argwhere(sum_cum > (np.random.rand()*sum_cum[-1])).flatten()[0]
                        w_t_k2_k1[k2, k1] = w_t_k2_k1[k2, k1] + 1
                        w_t_k2_v[k2, v] = w_t_k2_v[k2, v] + 1
        self._hyper_params.DirBN_para[t+1]['n_topic_word'] = w_t_k2_v
        self._hyper_params.DirBN_para[t+1]['n_topic_topic'] = w_t_k2_k1

    def _sample_Theta(self, theta_count=None):
        '''theta_count:  [numpy.ndarray]'''
        b0 = 0.01
        a0 = 0.01
        t = self._crt_sum_mex_matrix_v1(theta_count.T, self._hyper_params.theta_para['r_k'].flatten())

        self._sample_rk(t, self._hyper_params.theta_para['r_k'], self._hyper_params.theta_para['p_j'], self._hyper_params.theta_para['gamma0'], self._hyper_params.theta_para['c0'])
        self._hyper_params.theta_para['p_j'] = self._sampler.beta(np.sum(theta_count, 0) + a0, np.sum(self._hyper_params.theta_para['r_k'], 0) + b0).reshape(1, -1)
        self._hyper_params.theta_para['theta'] = self._sampler.gamma(self._hyper_params.theta_para['r_k'] + theta_count) * self._hyper_params.theta_para['p_j']

    def _sample_rk(self, XTplusOne_sum=None, r_k=None, p_jTplusOne=None, gamma0=None, c0=None, IsNoSample=None, e0=None,
                   f0=None, a0=None, b0=None):
        p_jTplusOne = p_jTplusOne.flatten()
        r_k = r_k.flatten()
        IsNoSample = False
        e0 = 1
        f0 = 1
        a0 = 0.01
        b0 = 0.01
        if (len(XTplusOne_sum.shape) > 1):
            if (XTplusOne_sum.shape[1] > 1):
                # XTplusOne_sum = np.full(np.sum(XTplusOne_sum, 1))
                raise('unexcepted error')

        KT = len(r_k)
        if not IsNoSample:
            c0 = self._sampler.gamma(e0 + gamma0) / (f0 + np.sum(r_k))
            sumlogpi = np.sum(np.log(np.maximum(1-p_jTplusOne, realmin)))
            p_prime = - sumlogpi / (c0 - sumlogpi)
            # L_k = full(sum(XTplusOne,2));
            # XTplusOne_sum
            gamma0 = self._sampler.gamma(a0 + self._crt_sum_mex_v1(XTplusOne_sum, (gamma0 / KT))) / (b0 - np.log(np.maximum(1 - p_prime, realmin)))
            r_k = self._sampler.gamma(gamma0 / KT + XTplusOne_sum) / (c0 - sumlogpi)
        else:
            raise('unexcepted error')
            # c0 = (e0 + gamma0) / (f0 + np.sum(r_k))
            # sumlogpi = np.sum(np.log(np.max(1 - p_jTplusOne, realmin)))
            # p_prime = - sumlogpi / (c0 - sumlogpi)
            # # L_k = full(sum(XTplusOne,2));
            # # XTplusOne_sum;
            # #!!!!!!!!!!!!
            # temp = np.dot(gamma0 / KT, np.sum(np.psi(XTplusOne_sum + gamma0 / KT) - np.psi(gamma0 / KT)))
            # # Sample_rk.m:35
            # gamma0 = (a0 + temp) / (b0 - np.log(np.max(1 - p_prime, realmin)))
            # self._hyper_params.Theta_r_k = (gamma0 / KT + XTplusOne_sum) / (c0 - sumlogpi)
        self._hyper_params.theta_para['r_k'] = r_k.reshape(-1, 1)
        self._hyper_params.theta_para['gamma0'] = gamma0
        self._hyper_params.theta_para['c0'] = c0

    def _crt_sum_mex_matrix_v1(self, X, r):
        '''
        X: sparse(theta_count.T) N*K，
        r: self._hyper_params.theta_para['r_k'].T 1*K 0.1
        '''
        k, n = np.shape(X)
        lsum = np.zeros(n).astype(int)
        maxx = 0
        Xsparse = sparse.csc_matrix(X)
        pr = Xsparse.data
        ir = Xsparse.indices
        jc = Xsparse.indptr

        for j in range(n):
            starting_row_index = jc[j]
            stopping_row_index = jc[j+1]
            if starting_row_index == stopping_row_index:
                continue
            else:
                for current_row_index in range(starting_row_index, stopping_row_index):
                    maxx = int(max(maxx, pr[current_row_index]))
                prob = np.zeros(maxx)
                for i in range(maxx):
                    prob[i] = r[j] / (r[j] + i)

                lsum[j] = 0
                for current_row_index in range(starting_row_index, stopping_row_index):
                    for i in range(int(pr[current_row_index])):
                        if np.random.rand() <= prob[i]:
                            lsum[j] += 1

        return lsum

    def _crt_sum_mex_v1(self, x, r):
        if len(x.shape) == 2:
            Lenx = x.shape[0] * x.shape[1]
        else:
            Lenx = x.shape[0]
        maxx = 0
        for i in range(Lenx):
            if maxx < x[i]:
                maxx = x[i]
        prob = np.zeros(maxx)
        for i in range(maxx):
            prob[i] = r/(r+i)
        ij = 0
        Lsum = 0
        for i in range(Lenx):
            for j in range(x[i]):
                if np.random.rand() < prob[j]:
                    Lsum += 1
        return Lsum






import os
import numpy as np
import ctypes

from ._pre_process import para_preprocess


def find_in_path(name, path):
    "Find a file in a search path"
    # adapted fom http://code.activestate.com/recipes/52224-find-a-file-given-a-search-path/
    for dir in path.split(os.pathsep):
        binpath = os.path.join(dir, name)
        if os.path.exists(binpath):
            return os.path.abspath(binpath)
    return None

def get_nvcc_path():
    # get nvcc path
    if 'CUDAHOME' in os.environ:
        home = os.environ['CUDAHOME']
        nvcc = os.path.join(home, 'bin', 'nvcc')
    else:
        # otherwise, search the PATH for NVCC
        default_path = os.path.join(os.sep, 'usr', 'local', 'cuda', 'bin')
        nvcc = find_in_path('nvcc', os.environ['PATH'] + os.pathsep + default_path)
        # if nvcc is None:
        #     raise EnvironmentError('The nvcc binary could not be '
        #                            'located in your $PATH. Either add it to your path, or set $CUDAHOME')
    return nvcc


class distribution_sampler_gpu(object):

    def __init__(self, system_type='Windows', seed=0):
        """
        The basic class for sampling distribution on cpu
        """
        super(distribution_sampler_gpu, self).__init__()

        self.system_type = system_type
        self.seed = seed

        # ------------------------------------------------ basic sampler ------------------------------------------
        if system_type == 'Windows':
            '''
            To compile CUDA C/C++ under Windows system, Visual Studio and CUDA should have been installed.
            This module has been tested under Visual Studio 2019(with MSVC v142 - VS 2019 C++ x64/x86 tools) and CUDA Toolkit 11.5.
            '''
            compact_path = os.path.dirname(__file__) + "\_compact\sampler_kernel.dll"
            if not os.path.exists(compact_path):
                nvcc_path = get_nvcc_path()
                try:
                    os.system(nvcc_path+' -o '+compact_path+' --shared '+compact_path[:-4]+'_win.cu')
                except:
                    os.system(r'nvcc -o '+'"'+compact_path+'"'+' --shared '+'"'+compact_path[:-4]+'_win.cu'+'"')
            dll = ctypes.cdll.LoadLibrary(compact_path)
        
        elif system_type == 'Linux':
            compact_path = os.path.dirname(__file__) + "/_compact/sampler_kernel.so"
            if not os.path.exists(compact_path):
                nvcc_path = get_nvcc_path()
                os.system(nvcc_path+' -Xcompiler -fPIC -shared -o '+compact_path+' '+compact_path[:-3]+'_linux.cu')
            dll = ctypes.cdll.LoadLibrary(compact_path)

        # ------------------------------------------------substorage ------------------------------------------
        self._init_status = dll._init_status
        self._init_status.argtypes = [ctypes.c_size_t]
        self._init_status.restype = ctypes.c_void_p
        self.rand_status = self._init_status(self.seed)

        # ----------------------------------------------cuda sampler ------------------------------------------
        self._sample_gamma = dll._sample_gamma
        self._sample_gamma.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_standard_gamma = dll._sample_standard_gamma
        self._sample_standard_gamma.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_beta = dll._sample_beta
        self._sample_beta.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_normal = dll._sample_normal
        self._sample_normal.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_standard_normal = dll._sample_standard_normal
        self._sample_standard_normal.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_void_p]

        self._sample_uniform = dll._sample_uniform
        self._sample_uniform.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_standard_uniform = dll._sample_standard_uniform
        self._sample_standard_uniform.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_void_p]

        # self._sample_binomial = dll._sample_binomial
        # self._sample_binomial.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_negative_binomial = dll._sample_negative_binomial
        self._sample_negative_binomial.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_multinomial = dll._sample_multinomial
        self._sample_multinomial.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_poisson = dll._sample_poisson
        self._sample_poisson.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_crt = dll._sample_crt
        self._sample_crt.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_cauchy = dll._sample_cauchy
        self._sample_cauchy.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_standard_cauchy = dll._sample_standard_cauchy
        self._sample_standard_cauchy.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_void_p]

        self._sample_chisquare = dll._sample_chisquare
        self._sample_chisquare.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_noncentral_chisquare = dll._sample_noncentral_chisquare
        self._sample_noncentral_chisquare.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_exponential = dll._sample_exponential
        self._sample_exponential.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_f = dll._sample_f
        self._sample_f.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_noncentral_f = dll._sample_noncentral_f
        self._sample_noncentral_f.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_geometric = dll._sample_geometric
        self._sample_geometric.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_gumbel = dll._sample_gumbel
        self._sample_gumbel.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_hypergeometric = dll._sample_hypergeometric
        self._sample_hypergeometric.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_laplace = dll._sample_laplace
        self._sample_laplace.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_logistic = dll._sample_logistic
        self._sample_logistic.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_power = dll._sample_power
        self._sample_power.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_zipf = dll._sample_zipf
        self._sample_zipf.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_pareto = dll._sample_pareto
        self._sample_pareto.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_rayleigh = dll._sample_rayleigh
        self._sample_rayleigh.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_t = dll._sample_t
        self._sample_t.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_triangular = dll._sample_triangular
        self._sample_triangular.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]

        self._sample_weibull = dll._sample_weibull
        self._sample_weibull.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p]


    def gamma(self, shape, scale=1.0, times=1):
        """
        sampler for the gamma distribution
        Inputs:
            shape  : [float] or [np.ndarray] shape parameter;
            scale  : [float] or [np.ndarray] scale parameter;
            times  : [int] the times required to sample;
            device : [str] 'cpu' or 'gpu';
        Outputs:
            output : [np.ndarray] or [pycuda.gpuarray] the resulting matrix on the device 'cpu' or 'gpu'
        """
        # in: non-negative
        matrix_scale, nElems, shape, scale, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, shape, scale)
        shape_p = ctypes.cast(shape.ctypes.data, ctypes.POINTER(ctypes.c_float))
        scale_p = ctypes.cast(scale.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_gamma(shape_p, scale_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def standard_gamma(self, shape, times=1):
        # shape: non-negative
        matrix_scale, nElems, shape, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, shape)
        shape_p = ctypes.cast(shape.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_standard_gamma(shape_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def dirichlet(self, shape, times=1):  # cacul by gamma. _sampler and cupy all take this way
        matrix_scale, nElems, shape, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, shape)
        shape_p = ctypes.cast(shape.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_standard_gamma(shape_p, output_p, matrix_scale, times, self.rand_status)
        if times == 1:
            output = output / np.sum(output, axis=-1, keepdims=True)
        else:
            output = output / np.sum(output, axis=-2, keepdims=True)
        return output[0] if scalar_flag else output

    def beta(self, a, b, times=1):
        # in: positive
        matrix_scale, nElems, a, b, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, a, b)
        a_p = ctypes.cast(a.ctypes.data, ctypes.POINTER(ctypes.c_float))
        b_p = ctypes.cast(b.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_beta(a_p, b_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def normal(self, loc=0.0, scale=1.0, times=1):
        # scale: non-negative
        matrix_scale, nElems, loc, scale, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, loc, scale)
        loc_p = ctypes.cast(loc.ctypes.data, ctypes.POINTER(ctypes.c_float))
        scale_p = ctypes.cast(scale.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_normal(loc_p, scale_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def standard_normal(self, size=1):
        assert type(size) == int or type(size) == tuple, "param size(Output shape) should be int or tuple of ints"
        nElems = size if type(size) == int else np.prod(size)
        output = np.empty(size, dtype=np.float32, order='C')
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_standard_normal(output_p, nElems, self.rand_status)
        return output[0] if size == 1 else output

    def uniform(self, low=0.0, high=1.0, times=1):
        # low < high
        matrix_scale, nElems, low, high, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, low, high)
        low_p = ctypes.cast(low.ctypes.data, ctypes.POINTER(ctypes.c_float))
        high_p = ctypes.cast(high.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_uniform(low_p, high_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def standard_uniform(self, size=1):
        assert type(size) == int or type(size) == tuple, "param size(Output shape) should be int or tuple of ints"
        nElems = size if type(size) == int else np.prod(size)
        output = np.empty(size, dtype=np.float32, order='C')
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_standard_uniform(output_p, nElems, self.rand_status)
        return output[0] if size == 1 else output

    def binomial(self, count=1, prob=0.5, times=1):  # 可以进行很深的优化cupy, 但依然转调multinomial
        """
        sampler for the gamma distribution
        Inputs:
            count  : [int] or [np.ndarray] count parameter;
            prob   : [float] or [np.ndarray] prob parameter, between 0 and 1;
            times  : [int] the times required to sample;
            device : [str] 'cpu' or 'gpu';
        Outputs:
            output : [np.ndarray] or [pycuda.gpuarray] the resulting matrix on the device 'cpu' or 'gpu'
        """
        count = np.array(count, dtype=np.int32, order='C')
        prob = np.array(prob, dtype=np.float32, order='C')

        assert count.size == prob.size, 'Param count and prob should have the same length.'
        assert len(count.shape) <= 2, 'Shape Error: the dimension of the input parameter a in the sampling distirbution shoud not be larger than 2'
        assert len(prob.shape) <= 2, 'Shape Error: the dimension of the input parameter b in the sampling distirbution shoud not be larger than 2'

        if count.size == prob.size:
            prob = np.expand_dims(prob, axis=-1)

        if prob.shape[-1] == 1:
            del_prob = 1 - prob
            prob = np.concatenate([prob, del_prob], axis=-1)
        return self.multinomial(count, prob, times)

    def negative_binomial(self, r, p, times=1):
        # r: *int, larger than 1
        # p: between 0 and 1
        # output: int
        matrix_scale, nElems, r, p, output, output_scale, scalar_flag = para_preprocess(times, [np.int32, np.float32], np.int32, r, p)
        r_p = ctypes.cast(r.ctypes.data, ctypes.POINTER(ctypes.c_int))
        p_p = ctypes.cast(p.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_int))

        self._sample_negative_binomial(r_p, p_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def multinomial(self, count=[1, 1], prob=[[0.5, 0.5], [0.2, 0.8]], times=1):
        """
        sampler for the multi distribution
        Inputs:
            count  : [int] or [np.ndarray] count parameter;
            prob   : [float] or [np.ndarray] prob parameter;
            times  : [int] the times required to sample;
            device : [str] 'cpu' or 'gpu';
        Outputs:
            output : [np.ndarray] or [pycuda.gpuarray] the resulting matrix on the device 'cpu' or 'gpu'
        """
        # in: non-negative
        # sum prob in a axis should be equal to 1
        count = np.array(count, dtype=np.int32, order='C')
        prob = np.array(prob, dtype=np.float32, order='C')

        assert count.size == 1 or count.size == prob.shape[0], 'Shape Error: '
        assert len(count.shape) <= 2, 'Shape Error: the dimension of the input parameter a in the sampling distirbution shoud not be larger than 2'
        assert len(prob.shape) <= 2, 'Shape Error: the dimension of the input parameter b in the sampling distirbution shoud not be larger than 2'

        output_scale = prob.shape if times == 1 or (times == 1 and count.size == 1) else prob.shape + (times,)
        output = np.zeros(output_scale, dtype=np.int32, order='C')
        matrix_scale_1 = count.size
        matrix_scale_2 = prob.size

        count_p = ctypes.cast(count.ctypes.data, ctypes.POINTER(ctypes.c_int))
        prob_p = ctypes.cast(prob.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_int))

        self._sample_multinomial(count_p, prob_p, output_p, matrix_scale_1, matrix_scale_2, times, self.rand_status)
        return output

    def poisson(self, lam=1.0, times=1):
        # lam: non-negative
        # output: int
        matrix_scale, nElems, lam, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.int32, lam)
        lam_p = ctypes.cast(lam.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_int))

        self._sample_poisson(lam_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def crt(self, customers, prob, times=1):  # dirichlet相近
        # Chinese restaurant process
        # https://qianyang-hfut.blog.csdn.net/article/details/52371443
        # customers 第几个顾客 prob开新比例 out桌子数量
        # customers: int
        # prob:
        # out: int
        matrix_scale, nElems, customers, prob, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.int32, customers, prob)
        customers_p = ctypes.cast(customers.ctypes.data, ctypes.POINTER(ctypes.c_float))
        prob_p = ctypes.cast(prob.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_int))

        self._sample_crt(customers_p, prob_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def cauchy(self, loc=0.0, scale=1.0, times=1):
        matrix_scale, nElems, loc, scale, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, loc, scale)
        loc_p = ctypes.cast(loc.ctypes.data, ctypes.POINTER(ctypes.c_float))
        scale_p = ctypes.cast(scale.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_cauchy(loc_p, scale_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def standard_cauchy(self, size=1):
        assert type(size) == int or type(size) == tuple, "param size(Output shape) should be int or tuple of ints"
        nElems = size if type(size) == int else np.prod(size)
        output = np.empty(size, dtype=np.float32, order='C')
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_standard_cauchy(output_p, nElems, self.rand_status)
        return output[0] if size == 1 else output

    def chisquare(self, degrees, times=1):
        # degrees int
        matrix_scale, nElems, degrees, output, output_scale, scalar_flag = para_preprocess(times, np.int32, np.float32, degrees)
        degrees_p = ctypes.cast(degrees.ctypes.data, ctypes.POINTER(ctypes.c_int))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_chisquare(degrees_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def noncentral_chisquare(self, df, nonc, times=1):
        # df: int
        # nonc: Non-centrality, must be non-negative.
        matrix_scale, nElems, df, nonc, output, output_scale, scalar_flag = para_preprocess(times, [np.int32, np.float32], np.float32, df, nonc)
        df_p = ctypes.cast(df.ctypes.data, ctypes.POINTER(ctypes.c_int))
        nonc_p = ctypes.cast(nonc.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_noncentral_chisquare(df_p, nonc_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def exponential(self, Lambda=1.0, times=1):
        matrix_scale, nElems, Lambda, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, Lambda)
        Lambda_p = ctypes.cast(Lambda.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_exponential(Lambda_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def standard_exponential(self, size=1):
        assert type(size) == int or type(size) == tuple, "param size(Output shape) should be int or tuple of ints"
        matrix_scale = size if type(size) == int else np.prod(size)
        Lambda = np.ones(size, dtype=np.float32, order='C')
        Lambda_p = ctypes.cast(Lambda.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output = np.empty(size, dtype=np.float32, order='C')
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_exponential(Lambda_p, output_p, matrix_scale, 1, self.rand_status)
        return output[0] if size == 1 else output

    def f(self, n1, n2, times=1):
        # n1, n2: int
        matrix_scale, nElems, n1, n2, output, output_scale, scalar_flag = para_preprocess(times, np.int32, np.float32, n1, n2)
        n1_p = ctypes.cast(n1.ctypes.data, ctypes.POINTER(ctypes.c_int))
        n2_p = ctypes.cast(n2.ctypes.data, ctypes.POINTER(ctypes.c_int))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_f(n1_p, n2_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def noncentral_f(self, dfnum, dfden, nonc, times=1):
        # dfnum, Numerator degrees of freedom
        # dfden, Denominator degrees of freedom, must be > 0.
        matrix_scale, nElems, dfnum, nonc, output, output_scale, scalar_flag = para_preprocess(times, [np.int32, np.float32], np.float32, dfnum, nonc)
        if type(dfden) in [float, int]:
            dfden = np.array([dfden] * matrix_scale, dtype=np.int32, order='C')
        else:
            dfden = np.array(dfden, dtype=np.int32, order='C')
            assert dfden.shape == dfnum.shape, 'param dfden should be scalar or have the same shape as dfnum or nonc.'

        dfnum_p = ctypes.cast(dfnum.ctypes.data, ctypes.POINTER(ctypes.c_int))
        dfden_p = ctypes.cast(dfden.ctypes.data, ctypes.POINTER(ctypes.c_int))
        nonc_p = ctypes.cast(nonc.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_noncentral_f(dfnum_p, dfden_p, nonc_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def geometric(self, p, times=1):
        # p: (0, 1)
        # output: int
        matrix_scale, nElems, p, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.int32, p)
        p_p = ctypes.cast(p.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_int))

        self._sample_geometric(p_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def gumbel(self, loc=0.0, scale=1.0, times=1):
        # scale: non-negative
        matrix_scale, nElems, loc, scale, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, loc, scale)
        loc_p = ctypes.cast(loc.ctypes.data, ctypes.POINTER(ctypes.c_float))
        scale_p = ctypes.cast(scale.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_gumbel(loc_p, scale_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def hypergeometric(self, ngood, nbad, nsample, times=1):
        # all input int
        # out int
        assert '等长?....'  # should be ngood.size >= nbad.size
        matrix_scale, nElems, ngood, nsample, output, output_scale, scalar_flag = para_preprocess(times, np.int32, np.int32, ngood, nsample)
        if type(nbad) in [float, int]:
            nbad = np.array([nbad] * matrix_scale, dtype=np.int32, order='C')
        else:
            nbad = np.array(nbad, dtype=np.int32, order='C')
            assert nbad.shape == ngood.shape, 'param nbad should be scalar or have the same shape as ngood or nsample.'
        ngood_p = ctypes.cast(ngood.ctypes.data, ctypes.POINTER(ctypes.c_int))
        nbad_p = ctypes.cast(nbad.ctypes.data, ctypes.POINTER(ctypes.c_int))
        nsample_p = ctypes.cast(nsample.ctypes.data, ctypes.POINTER(ctypes.c_int))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_int))

        self._sample_hypergeometric(ngood_p, nbad_p, nsample_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def laplace(self, loc=0.0, scale=1.0, times=1):
        # scale: non-negative
        matrix_scale, nElems, loc, scale, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, loc, scale)
        loc_p = ctypes.cast(loc.ctypes.data, ctypes.POINTER(ctypes.c_float))
        scale_p = ctypes.cast(scale.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_laplace(loc_p, scale_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def logistic(self, loc=0.0, scale=1.0, times=1):
        # scale: non-negative
        matrix_scale, nElems, loc, scale, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, loc, scale)
        loc_p = ctypes.cast(loc.ctypes.data, ctypes.POINTER(ctypes.c_float))
        scale_p = ctypes.cast(scale.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_logistic(loc_p, scale_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def power(self, a, times=1):
        # a: non-negative
        matrix_scale, nElems, a, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, a)
        a_p = ctypes.cast(a.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_power(a_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def zipf(self, a, times=1):
        # a: > 1
        # output: int
        matrix_scale, nElems, a, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.int32, a)
        a_p = ctypes.cast(a.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_int))

        self._sample_zipf(a_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def pareto(self, k, xm=1, times=1):
        # k > 1 (sampler, why?) 幂级数
        matrix_scale, nElems, k, xm, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, k, xm)
        k_p = ctypes.cast(k.ctypes.data, ctypes.POINTER(ctypes.c_float))
        xm_p = ctypes.cast(xm.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_pareto(k_p, xm_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def rayleigh(self, scale=1.0, times=1):
        # scale: non-negative
        matrix_scale, nElems, scale, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, scale)
        scale_p = ctypes.cast(scale.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_rayleigh(scale_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def t(self, df, times=1):
        # df: positive
        matrix_scale, nElems, df, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, df)
        df_p = ctypes.cast(df.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_t(df_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def triangular(self, left, mode, right, times=1):
        # 等长
        # left <= mode <= right
        # left < right
        assert '等长?....'
        matrix_scale, nElems, left, mode, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, left, mode)
        if type(right) in [float, int]:
            right = np.array([right]*matrix_scale, dtype=np.float32, order='C')
        else:
            right = np.array(right, dtype=np.float32, order='C')
            assert right.shape == left.shape, 'param right should be scalar or have the same shape as left or mode.'
        left_p = ctypes.cast(left.ctypes.data, ctypes.POINTER(ctypes.c_float))
        mode_p = ctypes.cast(mode.ctypes.data, ctypes.POINTER(ctypes.c_float))
        right_p = ctypes.cast(right.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_triangular(left_p, mode_p, right_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output

    def weibull(self, shape, scale, times=1):
        # a: non-negative (np. said)
        matrix_scale, nElems, shape, scale, output, output_scale, scalar_flag = para_preprocess(times, np.float32, np.float32, shape, scale)
        shape_p = ctypes.cast(shape.ctypes.data, ctypes.POINTER(ctypes.c_float))
        scale_p = ctypes.cast(scale.ctypes.data, ctypes.POINTER(ctypes.c_float))
        output_p = ctypes.cast(output.ctypes.data, ctypes.POINTER(ctypes.c_float))

        self._sample_weibull(shape_p, scale_p, output_p, matrix_scale, times, self.rand_status)
        return output[0] if scalar_flag else output






"""
===========================================
Metric
===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu
# License: BSD-3-Clause

import numpy as np

from scipy.special import gamma
from .._utils import *

def Poisson_Likelihood(X, X_re):

    # X[np.where(X>100)] = 100
    # X_re[np.where(X>100)] = 100

    Likelihood = np.sum(X*log_max(X_re) - X_re - log_max(gamma(X_re + 1)))
    return Likelihood

def Reconstruct_Error(X, X_re):
    return np.power(X - X_re, 2).sum()



# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Claus

class Params(object):
    def __init__(self):
        """
        The basic class for storing the parameters in the probabilistic model
        """
        super(Params, self).__init__()


class Basic_Model(object):
    def __init__(self, *args, **kwargs):
        """
        The basic model for all probabilistic models in this package
        Attributes:
            @public:
                global_params : [Params] the global parameters of the probabilistic model
                local_params  : [Params] the local parameters of the probabilistic model

            @private:
                _model_setting : [Params] the model settings of the probabilistic model
                _hyper_params  : [Params] the hyper parameters of the probabilistic model

        """
        super(Basic_Model, self).__init__()

        setattr(self, 'global_params', Params())
        setattr(self, 'local_params', Params())

        setattr(self, '_model_setting', Params())
        setattr(self, '_hyper_params', Params())




"""
===========================================
Poisson Factor Analysis
Beta-Negative Binomial Process and Poisson Factor Analysis
Mingyuan Zhou, Lauren Hannah, David Dunson, Lawrence Carin
Publihsed in International Conference on Artificial Intelligence and Statistic 2012

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>； Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import numpy as np
import scipy.io as sio

from pydpm._metric import ACC
from pydpm._model import PFA

# load data
data = sio.loadmat('./data/mnist_gray')
train_data = np.array(np.ceil(data['train_mnist']*5), order='C')[:, 0:999]
test_data = np.array(np.ceil(data['train_mnist']*5), order='C')[:, 1000:1999]
train_label = data['train_label'][:999]
test_label = data['train_label'][1000:1999]

# create the model and deploy it on gpu or cpu
model = PFA(128, 'gpu')
model.initial(train_data)  # use the shape of train_data to initialize the params of model
train_local_params = model.train(100, train_data)
train_local_params = model.test(100, train_data)
test_local_params = model.test(100, test_data)

# evaluate the model with classification accuracy
# the demo accuracy can achieve 0.8238
results = ACC(train_local_params.Theta, test_local_params.Theta, train_label, test_label, 'SVM')

# save the model after training
model.save()
# model.load('./save_models/PFA.npy')



import numpy as np
import time
import ctypes


def para_preprocess(times=1, in_type=np.float32, out_type=np.float32, *args):
    """
    preprocess the input parameters in sampling the distribution with gpu
    Inputs:
        times   : [int] repeat times
        in_type : [np.dtype] or list of np.dtype the dtype of the input parameters
        out_type: [np.dtype] the dtype of the output sampling results
        args[0] : [np.ndarray] the first variable in the target distribution
        args[1] : [np.ndarray] the second variable in the target distribution
    Outputs:
        para_a      : [pycuda.gpuarray] the input matrix for the first parameter
        para_b      : [pycuda.gpuarray] the input matrix for the second parameter
        output      : [pycuda.gpuarray] the matrix on gpu to store the sampling result
        para_scale  : [list] a list including the number of element and repeat times in the resulting matrix
        para_seed   : [pycuda.gpuarray] seed matrix on gpu
        partition   : [list] a list including
        scalar_flag : [bool] if the resulting matrix is a scalar
    """
    assert len(args) <= 2, 'Value Error: the number of the input parameter in the sampling distribution should not be larger than 2'

    if len(args) == 1:
        para_a = np.array(args[0], dtype=in_type, order='C')
        # assert len(para_a.shape) <= 2, 'Shape Error: the dimension of the input parameter a in the sampling distribution shoud not be larger than 2'

        # obtain the output_scale and judge if the para_a is a scalar
        if times > 1:
            output_scale = para_a.shape + (times,)
            scalar_flag = False
        else:
            output_scale = (1,) if para_a.shape == () else para_a.shape
            scalar_flag = True if para_a.shape == () else False

    elif len(args) == 2:
        if type(in_type) == type:
            in_type = [in_type, in_type]
        assert (type(in_type) == list and len(in_type) == 2)

        para_a = np.array(args[0], dtype=in_type[0], order='C')
        para_b = np.array(args[1], dtype=in_type[1], order='C')
        # assert len(para_a.shape) <= 2, 'Shape Error: the dimension of the input parameter a in the sampling distirbution shoud not be larger than 2'
        # assert len(para_b.shape) <= 2, 'Shape Error: the dimension of the input parameter b in the sampling distirbution shoud not be larger than 2'

        # make sure the sizes of para_a and para_b are equal
        if para_a.size == 1 and para_b.size != 1:
            para_a = np.array(np.full(para_b.shape, para_a), dtype=in_type[0], order='C')
        if para_b.size == 1 and para_a.size != 1:
            para_b = np.array(np.full(para_a.shape, para_b), dtype=in_type[1], order='C')

        # obtain the output_scale and judge if the para_a is a scalar
        if times > 1:
            output_scale = para_a.shape + (times,)
            scalar_flag = False
        else:
            output_scale = (1,) if para_a.shape == () else para_a.shape
            scalar_flag = True if para_a.shape == () else False

    matrix_scale = para_a.size
    nElems = para_a.size * times  # output_scale multi.

    # output
    output = np.empty(output_scale, dtype=out_type, order='C')

    if len(args) == 1:
        return matrix_scale, nElems, para_a, output, output_scale, scalar_flag
    elif len(args) == 2:
        return matrix_scale, nElems, para_a, para_b, output, output_scale, scalar_flag


"""
===========================================
Multimodal Poisson Gamma Belief Network
Chaojie Wang, Bo Chen and Mingyuan Zhou
Published in In AAAI Conference on Artificial Intelligence

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Jiawen Wu <wjw19960807@163.com>; Wei Zhao <13279389260@163.com>
# License: BSD-3-Clause

import numpy as np
import scipy.io as sio

from pydpm._metric import ACC
from pydpm._model import MPGBN

# load data
data = sio.loadmat('./data/mnist_gray')
train_data = np.array(np.ceil(data['train_mnist']*5), order='C')[:, 0:999]
train_data_1 = train_data[:360, :]
train_data_2 = train_data[360:, :]

test_data = np.array(np.ceil(data['train_mnist']*5), order='C')[:, 1000:1999]
test_data_1 = test_data[:360, :]
test_data_2 = test_data[360:, :]

train_label = data['train_label'][:999]
test_label = data['train_label'][1000:1999]


# create the model and deploy it on gpu or cpu
model = MPGBN([128, 64, 32], device='gpu')
model.initial(train_data_1, train_data_2)
train_local_params = model.train(100, train_data_1, train_data_2)
train_local_params = model.test(100, train_data_1, train_data_2)
test_local_params = model.test(100, test_data_1, test_data_2)

# evaluate the model with classification accuracy
# the demo accuracy can achieve 0.8549
results = ACC(train_local_params.Theta[0], test_local_params.Theta[0], train_label, test_label, 'SVM')

# save the model after training
model.save()
# model.load('./save_models/PGBN.npy')





"""
===========================================
Deep Relational Topic Modeling via Graph Poisson Gamma Belief Network
Chaojie Wang, Hao Zhang, Bo Chen, Dongsheng Wang, Zhengjue Wang
Published in Advances in Neural Information Processing System 2020

===========================================

"""

# Author: Chaojie Wang <xd_silly@163.com>; Wei Zhao <13279389260@163.com>; Jiawen Wu <wjw19960807@163.com>
# License: BSD-3-Clause

import numpy as np
import scipy.io as sio

from pydpm._metric import ACC
from pydpm._model import GPGBN
from pydpm._utils import cosine_simlarity

# load data
data = sio.loadmat('./data/mnist_gray')
train_data = np.array(np.ceil(data['train_mnist']*25), order='C')[:, 0:50]
train_label = data['train_label'][:50]

# construct the adjacency matrix
train_graph = cosine_simlarity(train_data.T, train_data.T)
train_graph[np.where(train_graph < 0.5)] = 0

# create the model and deploy it on gpu or cpu
model = GPGBN([128, 64, 32], device='gpu')
model.initial(train_data)
train_local_params = model.train(500, train_data, train_graph)

# save the model after training
model.save()
# model.load('./save_models/PGBN.npy')


